/*  Misfit Model 3D
 * 
 *  Copyright (c) 2004 Kevin Worcester
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, 
 *  USA.
 *
 *  See the COPYING file for full license text.
 */


#include "model.h"
#include "modelundo.h"
#include "log.h"
#include "texture.h"
#include "texmgr.h"
#include "modelstatus.h"

#include <GL/glu.h>

int Model::s_glTextures = 0;

bool Model::loadTextures( unsigned context )
{
   LOG_PROFILE();

   TextureContext * texContext = NULL;
   if ( context )
   {
      deleteGlTextures( context );
      texContext = getTextureContext( context );
      texContext->m_textures.clear();
   }

   for ( unsigned t = 0; t < m_materials.size(); t++ )
   {
      if ( texContext )
      {
         texContext->m_textures.push_back( -1 );
      }

      if ( m_materials[t]->m_filename[0] 
            && m_materials[t]->m_type == Model::Material::MATTYPE_TEXTURE )
      {
         Texture * tex = TextureManager::getInstance()->getTexture( m_materials[t]->m_filename.c_str() );

         if ( !tex )
         {
            model_status( this, StatusError, STATUSTIME_LONG, "Could not load texture %s", m_materials[t]->m_filename.c_str() );
            tex = TextureManager::getInstance()->getDefaultTexture( m_materials[t]->m_filename.c_str() );
         }
         
         if ( tex )
         {
            m_materials[t]->m_textureData = tex;

            glGenTextures( 1, &(m_materials[t]->m_texture) );
            s_glTextures++;

            if ( texContext )
            {
               texContext->m_textures[t] = m_materials[t]->m_texture;
            }

            log_debug( "loaded texture %s as %d\n", tex->m_name, m_materials[t]->m_texture );

            if ( context )
            {
               glBindTexture( GL_TEXTURE_2D, m_materials[t]->m_texture );

               glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER,
                     GL_LINEAR );
               glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER,
                     GL_LINEAR );

               GLuint format = tex->m_format == Texture::FORMAT_RGBA ? GL_RGBA : GL_RGB;
               gluBuild2DMipmaps( GL_TEXTURE_2D, format,
                     tex->m_width, tex->m_height,
                     format, GL_UNSIGNED_BYTE,
                     tex->m_data );
            }
         }
         else
         {
            log_error( "Could not load texture %s\n", m_materials[t]->m_filename.c_str() );
         }
      }
   }

   if ( texContext )
   {
      texContext->m_valid = true;
   }

   return true;
}

int Model::addTexture( Texture * tex )
{
   LOG_PROFILE();
   if ( m_animationMode )
   {
      return -1;
   }


   if ( tex )
   {
      int num = m_materials.size();

      Material * material = Material::get();

      material->m_name = tex->m_name;
      material->m_texture = 0;
      material->m_textureData = tex;
      material->m_filename = tex->m_filename;
      for ( int m = 0; m < 3; m++ )
      {
         material->m_ambient[m] = 0.2;
         material->m_diffuse[m] = 0.8;
         material->m_specular[m] = 0.0;
         material->m_emissive[m] = 0.0;
      }
      material->m_ambient[3]  = 1.0;
      material->m_diffuse[3]  = 1.0;
      material->m_specular[3] = 1.0;
      material->m_emissive[3] = 1.0;

      material->m_shininess = 0.0;

      //TextureContextList m_textureContexts;
      m_materials.push_back( material );

      MU_AddTexture * undo = new MU_AddTexture();
      undo->addTexture( num, material );
      sendUndo( undo );

      invalidateTextures();
      return num;
   }
   else
   {
      return -1;
   }
}

void Model::deleteTexture( unsigned textureNum )
{
   LOG_PROFILE();
   if ( m_animationMode )
   {
      return;
   }


   for ( unsigned g = 0; g < m_groups.size(); g++ )
   {
      if ( m_groups[g]->m_materialIndex == (signed) textureNum )
      {
         setGroupTextureId( g, -1 );
      }
      if ( m_groups[g]->m_materialIndex > (signed) textureNum )
      {
         setGroupTextureId( g, m_groups[g]->m_materialIndex - 1 );
      }
   }

   MU_DeleteTexture * undo = new MU_DeleteTexture();
   undo->deleteTexture( textureNum, m_materials[ textureNum ] );
   sendUndo( undo );

   removeTexture( textureNum );
}

int Model::getGroupTextureId( unsigned groupNumber )
{
   if ( groupNumber >= 0 && groupNumber < m_groups.size() )
   {
      return m_groups[ groupNumber ]->m_materialIndex;
   }
   else
   {
      return -1;
   }
}

bool Model::setGroupTextureId( unsigned groupNumber, int textureId )
{
   if ( m_animationMode )
   {
      return false;
   }

   if ( groupNumber >= 0 && groupNumber < m_groups.size() && textureId < (int) m_materials.size() )
   {
      MU_SetTexture * undo = new MU_SetTexture();
      undo->setTexture( groupNumber, textureId,
            m_groups[ groupNumber ]->m_materialIndex );
      sendUndo( undo );

      m_groups[ groupNumber ]->m_materialIndex = textureId;
      return true;
   }
   else
   {
      return false;
   }
}

bool Model::getTextureCoords( const unsigned & triangleNumber, const unsigned & vertexIndex, float & s, float & t )
{
   if ( triangleNumber < m_triangles.size() && vertexIndex < 3 )
   {
      s = m_triangles[ triangleNumber ]->m_s[ vertexIndex ];
      t = m_triangles[ triangleNumber ]->m_t[ vertexIndex ];
      return true;
   }
   else
   {
      return false;
   }
}

bool Model::setTextureCoords( const unsigned & triangleNumber, const unsigned & vertexIndex, const float & s, const float & t )
{
   if ( m_animationMode )
   {
      return false;
   }

   //log_debug( "setTextureCoords( %d, %d, %f, %f )\n", triangleNumber, vertexIndex, s, t );

   if ( triangleNumber < m_triangles.size() && vertexIndex < 3 )
   {
      MU_SetTextureCoords * undo = new MU_SetTextureCoords();
      undo->addTextureCoords( triangleNumber, vertexIndex, s, t,
            m_triangles[ triangleNumber ]->m_s[ vertexIndex ], 
            m_triangles[ triangleNumber ]->m_t[ vertexIndex ] );
      sendUndo( undo );

      m_triangles[ triangleNumber ]->m_s[ vertexIndex ] = s;
      m_triangles[ triangleNumber ]->m_t[ vertexIndex ] = t;
      return true;
   }
   else
   {
      return false;
   }
}

Texture * Model::getTextureData( unsigned textureId )
{
   if ( textureId >= 0 && textureId < m_materials.size() )
   {
      return m_materials[textureId]->m_textureData;
   }
   else
   {
      return NULL;
   }
}

bool Model::getTextureAmbient( unsigned textureId, float * ambient )
{
   if ( ambient && textureId < m_materials.size() )
   {
      for ( int t = 0; t < 4; t++ )
      {
         ambient[t] = m_materials[textureId]->m_ambient[t];
      }
      return true;
   }
   else
   {
      return false;
   }
}

bool Model::setTextureAmbient( unsigned textureId, const float * ambient )
{
   if ( m_animationMode )
   {
      return false;
   }

   if ( ambient && textureId < m_materials.size() )
   {
      MU_SetLightProperties * undo = new MU_SetLightProperties();
      undo->setLightProperties( textureId, 
            MU_SetLightProperties::LightAmbient, 
            ambient, m_materials[textureId]->m_ambient);
      sendUndo( undo );

      for ( int t = 0; t < 4; t++ )
      {
         m_materials[textureId]->m_ambient[t] = ambient[t];
      }
      return true;
   }
   else
   {
      return false;
   }
}

bool Model::getTextureDiffuse( unsigned textureId, float * diffuse )
{
   if ( diffuse && textureId < m_materials.size() )
   {
      for ( int t = 0; t < 4; t++ )
      {
         diffuse[t] = m_materials[textureId]->m_diffuse[t];
      }
      return true;
   }
   else
   {
      return false;
   }
}

bool Model::setTextureDiffuse( unsigned textureId, const float * diffuse )
{
   if ( m_animationMode )
   {
      return false;
   }

   if ( diffuse && textureId < m_materials.size() )
   {
      MU_SetLightProperties * undo = new MU_SetLightProperties();
      undo->setLightProperties( textureId, 
            MU_SetLightProperties::LightDiffuse, 
            diffuse, m_materials[textureId]->m_diffuse);
      sendUndo( undo );

      for ( int t = 0; t < 4; t++ )
      {
         m_materials[textureId]->m_diffuse[t] = diffuse[t];
      }
      return true;
   }
   else
   {
      return false;
   }
}

bool Model::getTextureSpecular( unsigned textureId, float * specular )
{
   if ( specular && textureId < m_materials.size() )
   {
      for ( int t = 0; t < 4; t++ )
      {
         specular[t] = m_materials[textureId]->m_specular[t];
      }
      return true;
   }
   else
   {
      return false;
   }
}

bool Model::setTextureSpecular( unsigned textureId, const float * specular )
{
   if ( m_animationMode )
   {
      return false;
   }

   if ( specular && textureId < m_materials.size() )
   {
      MU_SetLightProperties * undo = new MU_SetLightProperties();
      undo->setLightProperties( textureId, 
            MU_SetLightProperties::LightSpecular, 
            specular, m_materials[textureId]->m_specular);
      sendUndo( undo );

      for ( int t = 0; t < 4; t++ )
      {
         m_materials[textureId]->m_specular[t] = specular[t];
      }
      return true;
   }
   else
   {
      return false;
   }
}

bool Model::getTextureEmissive( unsigned textureId, float * emissive )
{
   if ( emissive && textureId < m_materials.size() )
   {
      for ( int t = 0; t < 4; t++ )
      {
         emissive[t] = m_materials[textureId]->m_emissive[t];
      }
      return true;
   }
   else
   {
      return false;
   }
}

bool Model::setTextureEmissive( unsigned textureId, const float * emissive )
{
   if ( m_animationMode )
   {
      return false;
   }

   if ( emissive && textureId < m_materials.size() )
   {
      MU_SetLightProperties * undo = new MU_SetLightProperties();
      undo->setLightProperties( textureId, 
            MU_SetLightProperties::LightEmissive, 
            emissive, m_materials[textureId]->m_emissive);
      sendUndo( undo );

      for ( int t = 0; t < 4; t++ )
      {
         m_materials[textureId]->m_emissive[t] = emissive[t];
      }
      return true;
   }
   else
   {
      return false;
   }
}

bool Model::getTextureShininess( unsigned textureId, float & shininess )
{
   if ( textureId < m_materials.size() )
   {
      shininess = m_materials[textureId]->m_shininess;
      return true;
   }
   else
   {
      return false;
   }
}

bool Model::setTextureShininess( unsigned textureId, const float & shininess )
{
   if ( m_animationMode )
   {
      return false;
   }

   if ( textureId < m_materials.size() )
   {
      MU_SetShininess * undo = new MU_SetShininess();
      undo->setShininess( textureId, shininess, m_materials[textureId]->m_shininess);
      sendUndo( undo );

      m_materials[textureId]->m_shininess = shininess;
      return true;
   }
   else
   {
      return false;
   }
}

void Model::setTextureName( unsigned textureId, const char * name )
{
   if ( m_animationMode )
   {
      return;
   }

   if ( name && textureId < m_materials.size() )
   {
      MU_SetTextureName * undo = new MU_SetTextureName();
      undo->setTextureName( textureId, name, m_materials[ textureId ]->m_name.c_str() );
      sendUndo( undo );

      m_materials[ textureId ]->m_name = name;
   }
}

bool Model::setTextureSClamp( unsigned textureId, bool clamp )
{
   return false;
}

bool Model::getTextureSClamp( unsigned textureId )
{
   return false;
}

bool Model::setTextureTClamp( unsigned textureId, bool clamp )
{
   return false;
}

bool Model::getTextureTClamp( unsigned textureId )
{
   return false;
}

const char * Model::getTextureName( unsigned textureId )
{
   if ( textureId >= 0 && textureId < m_materials.size() )
   {
      return m_materials[textureId]->m_name.c_str();
   }
   else
   {
      return NULL;
   }
}

const char * Model::getTextureFilename( unsigned textureId )
{
   if ( textureId >= 0 && textureId < m_materials.size() )
   {
      return m_materials[textureId]->m_filename.c_str();
   }
   else
   {
      return NULL;
   }
}

int Model::getMaterialByName( const char * const materialName, bool ignoreCase )
{
   int (*compare)(const char *, const char *);
   compare = ignoreCase ? strcasecmp : strcmp;

   int matNumber = -1;

   for ( unsigned m = 0; m < m_materials.size(); m++ )
   {
      if ( compare( materialName, m_materials[m]->m_name.c_str() ) == 0 )
      {
         matNumber = m;
         break;
      }
   }

   return matNumber;
}

Model::TextureContext * Model::getTextureContext( unsigned context )
{
   TextureContextList::iterator it;
   for ( it = m_textureContexts.begin(); it != m_textureContexts.end(); it++ )
   {
      if ( (*it)->m_context == context )
      {
         return (*it);
      }
   }

   TextureContext * texContext = new TextureContext;
   texContext->m_context = context;
   texContext->m_valid   = false;
   m_textureContexts.push_back( texContext );

   return texContext;
}

void Model::invalidateTextures()
{
   TextureContextList::iterator it;
   for ( it = m_textureContexts.begin(); it != m_textureContexts.end(); it++ )
   {
      (*it)->m_valid = false;
   }
}

void Model::deleteGlTextures( unsigned context )
{
   TextureContext * texContext = getTextureContext( context );
   for ( unsigned t = 0; t < texContext->m_textures.size(); t++ )
   {
      int texId = texContext->m_textures[t];
      if ( texId > 0 )
      {
         glDeleteTextures( 1, (GLuint *) &texId );
         s_glTextures--;
      }
      texContext->m_textures[t] = -1;
   }
   texContext->m_valid = false;
}

void Model::noTexture( unsigned id )
{
   LOG_PROFILE();

   for ( unsigned t = 0; t < m_groups.size(); t++ )
   {
      if ( (unsigned) m_groups[t]->m_materialIndex == id )
      {
         m_groups[t]->m_materialIndex = -1;
      }
   }
}

void Model::removeContext( unsigned context )
{
   TextureContext * texContext = NULL;
   TextureContextList::iterator it;
   for ( it = m_textureContexts.begin(); texContext == NULL && it != m_textureContexts.end(); it++ )
   {
      if ( (*it)->m_context == context )
      {
         texContext = *it;
         deleteGlTextures( context );
         break;
      }
   }

   if ( texContext )
   {
      m_textureContexts.erase( it );
      delete texContext;
   }
}
