/*  Misfit Model 3D
 * 
 *  Copyright (c) 2004 Kevin Worcester
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, 
 *  USA.
 *
 *  See the COPYING file for full license text.
 */


#include "modelundo.h"
#include "model.h"

#include "log.h"

int ModelUndo::s_allocated = 0;

// Well, I suppose you can call it an implementation...
bool MU_NoOp::combine( Undo * u )
{
   if ( dynamic_cast<MU_NoOp *>( u ) )
   {
      return true;
   }
   else
   {
      log_debug( "couldn't combine with NoOp\n" );
      return false;
   }
}

MU_TranslateSelectedVertices::MU_TranslateSelectedVertices()
{
}

MU_TranslateSelectedVertices::~MU_TranslateSelectedVertices()
{
}

bool MU_TranslateSelectedVertices::combine( Undo * u )
{
   MU_TranslateSelectedVertices * undo = dynamic_cast<MU_TranslateSelectedVertices *>( u );

   if ( undo )
   {
      m_matrix = m_matrix * undo->m_matrix;
      return true;
   }
   else
   {
      log_debug( "couldn't combine with translate vertices\n" );
      return false;
   }
}

void MU_TranslateSelectedVertices::setMatrix( const Matrix & rhs )
{
   m_matrix = rhs;
}

void MU_TranslateSelectedVertices::undo( Model * model )
{
   log_debug( "undo translate selected vertices\n" );

   Matrix m = m_matrix.getInverse();

   /*
   // Invert our translation matrix
   for ( int c = 0; c < 3; c++ )
   {
      m.set( 3, c, -m_matrix.get( 3, c ) );
   }
   */

   model->translateSelectedVertices( m );
}

void MU_TranslateSelectedVertices::redo( Model * model )
{
   model->translateSelectedVertices( m_matrix );
}

MU_RotateSelectedVerticesOnPoint::MU_RotateSelectedVerticesOnPoint()
{
}

MU_RotateSelectedVerticesOnPoint::~MU_RotateSelectedVerticesOnPoint()
{
}

bool MU_RotateSelectedVerticesOnPoint::combine( Undo * u )
{
   MU_RotateSelectedVerticesOnPoint * undo = dynamic_cast<MU_RotateSelectedVerticesOnPoint *>( u );

   if ( undo )
   {
      for ( int t = 0; t < 3; t++ )
      {
         if ( m_point[t] != undo->m_point[t] )
         {
            log_debug( "couldn't combine with rotate vertices, points dissimilar\n" );
            return false;
         }
      }

      m_matrix = m_matrix * undo->m_matrix;
      return true;
   }
   else
   {
      log_debug( "couldn't combine with rotate vertices\n" );
      return false;
   }
}

void MU_RotateSelectedVerticesOnPoint::setMatrixPoint( const Matrix & rhs, double * point )
{
   m_matrix = rhs;
   for ( int t = 0; t < 3; t++ )
   {
      m_point[t] = point[t];
   }
}

void MU_RotateSelectedVerticesOnPoint::undo( Model * model )
{
   log_debug( "undo rotate selected vertices on point\n" );
   Matrix inv = m_matrix.getInverse();
   model->rotateSelectedVerticesOnPoint( inv, m_point );
}

void MU_RotateSelectedVerticesOnPoint::redo( Model * model )
{
   log_debug( "undo rotate selected vertices on point\n" );
   model->rotateSelectedVerticesOnPoint( m_matrix, m_point );
}

MU_ApplyMatrix::MU_ApplyMatrix()
{
}

MU_ApplyMatrix::~MU_ApplyMatrix()
{
}

bool MU_ApplyMatrix::combine( Undo * u )
{
   MU_ApplyMatrix * undo = dynamic_cast<MU_ApplyMatrix *>( u );

   if ( undo )
   {
      m_matrix = m_matrix * undo->m_matrix;
      return true;
   }
   else
   {
      log_debug( "couldn't combine with rotate vertices\n" );
      return false;
   }
}

void MU_ApplyMatrix::setMatrix( const Matrix & rhs )
{
   m_matrix = rhs;
}

void MU_ApplyMatrix::undo( Model * model )
{
   log_debug( "undo apply matrix\n" );
   Matrix m;

   m = m_matrix.getInverse();

   model->applyMatrix( m );
}

void MU_ApplyMatrix::redo( Model * model )
{
   model->applyMatrix( m_matrix );
}

MU_SelectionMode::MU_SelectionMode()
{
}

MU_SelectionMode::~MU_SelectionMode()
{
}

void MU_SelectionMode::undo( Model * model )
{
   log_debug( "undo selection mode %d\n", m_oldMode );

   model->setSelectionMode( m_oldMode );
}

void MU_SelectionMode::redo( Model * model )
{
   log_debug( "redo selection mode %d\n", m_mode );

   model->setSelectionMode( m_mode );
}

bool MU_SelectionMode::combine( Undo * u )
{
   MU_SelectionMode * undo = dynamic_cast< MU_SelectionMode * >( u );
   if ( undo )
   {
      m_mode = undo->m_mode;
      return true;
   }
   else
   {
      log_debug( "couldn't combine with select\n" );
      return false;
   }
}

unsigned MU_SelectionMode::size()
{
   return sizeof(MU_SelectionMode);
}

MU_Select::MU_Select( Model::SelectionMode mode )
   : m_mode( mode )
{
}

MU_Select::~MU_Select()
{
}

void MU_Select::undo( Model * model )
{
   SelectionDifferenceList::iterator it;

   // Invert selection from our list
   for ( it = m_diff.begin(); it != m_diff.end(); it++ )
   {
      if ( !(*it).oldSelected )
      {
         switch ( m_mode )
         {
            case Model::SelectVertices:
               model->unselectVertex( (*it).number );
               break;
            case Model::SelectTriangles:
               model->unselectTriangle( (*it).number );
               break;
            case Model::SelectGroups:
               model->unselectGroup( (*it).number );
               break;
            case Model::SelectJoints:
               model->unselectBoneJoint( (*it).number );
               break;
            case Model::SelectNone:
            default:
               break;
         }
      }
      else
      {
         switch ( m_mode )
         {
            case Model::SelectVertices:
               model->selectVertex( (*it).number );
               break;
            case Model::SelectTriangles:
               model->selectTriangle( (*it).number );
               break;
            case Model::SelectGroups:
               model->selectGroup( (*it).number );
               break;
            case Model::SelectJoints:
               model->selectBoneJoint( (*it).number );
               break;
            case Model::SelectNone:
            default:
               break;
         }
      }
   }
}

void MU_Select::redo( Model * model )
{
   log_debug( "redo select %d\n", m_mode );

   SelectionDifferenceList::iterator it;

   // Set selection from our list
   if ( m_diff.size() )
   {
      for ( it = m_diff.begin(); it != m_diff.end(); it++ )
      {
         if ( (*it).selected )
         {
            switch ( m_mode )
            {
               case Model::SelectVertices:
                  model->selectVertex( (*it).number );
                  break;
               case Model::SelectTriangles:
                  model->selectTriangle( (*it).number );
                  break;
               case Model::SelectGroups:
                  model->selectGroup( (*it).number );
                  break;
               case Model::SelectJoints:
                  model->selectBoneJoint( (*it).number );
                  break;
               case Model::SelectNone:
               default:
                  break;
            }
         }
         else
         {
            switch ( m_mode )
            {
               case Model::SelectVertices:
                  model->unselectVertex( (*it).number );
                  break;
               case Model::SelectTriangles:
                  model->unselectTriangle( (*it).number );
                  break;
               case Model::SelectGroups:
                  model->unselectGroup( (*it).number );
                  break;
               case Model::SelectJoints:
                  model->unselectBoneJoint( (*it).number );
                  break;
               case Model::SelectNone:
               default:
                  break;
            }
         }
      }
   }
}

bool MU_Select::combine( Undo * u )
{
   /*
   MU_Select * undo = dynamic_cast< MU_Select * >( u );
   if ( undo && getSelectionMode() == undo->getSelectionMode() )
   {
      SelectionDifferenceList::iterator it;
      for ( it = undo->m_diff.begin(); it != undo->m_diff.end(); it++ )
      {
         setSelectionDifference( (*it).number, (*it).selected, (*it).oldSelected );
      }
      return true;
   }
   else
   */
   {
      log_debug( "couldn't combine with select\n" );
      return false;
   }
}

unsigned MU_Select::size()
{
   return sizeof(MU_Select) + m_diff.size() * sizeof(SelectionDifference);
}

void MU_Select::setSelectionDifference( int number, bool selected, bool oldSelected )
{
   unsigned index;
   SelectionDifference diff;
   diff.number   = number;
   // Change selection state if it exists in our list
   if ( m_diff.find_sorted( diff, index ) )
   {
      m_diff[index].selected = selected;
   }

   // add selection state to our list
   diff.selected = selected;
   diff.oldSelected = oldSelected;

   m_diff.insert_sorted( diff );
}

MU_Hide::MU_Hide( Model::SelectionMode mode )
   : m_mode( mode )
{
}

MU_Hide::~MU_Hide()
{
}

void MU_Hide::undo( Model * model )
{
   log_debug( "undo hide\n" );

   HideDifferenceList::iterator it;

   // Invert visible from our list
   for ( it = m_diff.begin(); it != m_diff.end(); it++ )
   {
      if ( (*it).visible )
      {
         switch ( m_mode )
         {
            case Model::SelectVertices:
               model->hideVertex( (*it).number );
               break;
            case Model::SelectTriangles:
               model->hideTriangle( (*it).number );
               break;
            case Model::SelectJoints:
               model->hideJoint( (*it).number );
               break;
            case Model::SelectNone:
            default:
               break;
         }
      }
      else
      {
         switch ( m_mode )
         {
            case Model::SelectVertices:
               model->unhideVertex( (*it).number );
               break;
            case Model::SelectTriangles:
               model->unhideTriangle( (*it).number );
               break;
            case Model::SelectJoints:
               model->unhideJoint( (*it).number );
               break;
            case Model::SelectNone:
            default:
               break;
         }
      }
   }
}

void MU_Hide::redo( Model * model )
{
   HideDifferenceList::iterator it;

   // Set visible from our list
   for ( it = m_diff.begin(); it != m_diff.end(); it++ )
   {
      if ( (*it).visible )
      {
         switch ( m_mode )
         {
            case Model::SelectVertices:
               model->unhideVertex( (*it).number );
               break;
            case Model::SelectTriangles:
               model->unhideTriangle( (*it).number );
               break;
            case Model::SelectJoints:
               model->unhideJoint( (*it).number );
               break;
            case Model::SelectNone:
            default:
               break;
         }
      }
      else
      {
         switch ( m_mode )
         {
            case Model::SelectVertices:
               model->hideVertex( (*it).number );
               break;
            case Model::SelectTriangles:
               model->hideTriangle( (*it).number );
               break;
            case Model::SelectJoints:
               model->hideJoint( (*it).number );
               break;
            case Model::SelectNone:
            default:
               break;
         }
      }
   }
}

bool MU_Hide::combine( Undo * u )
{
   MU_Hide * undo = dynamic_cast< MU_Hide * >( u );

   if ( undo && getSelectionMode() == undo->getSelectionMode() )
   {
      HideDifferenceList::iterator it;
      for ( it = undo->m_diff.begin(); it != undo->m_diff.end(); it++ )
      {
         setHideDifference( (*it).number, (*it).visible );
      }
      return true;
   }
   else
   {
      log_debug( "couldn't combine with hide\n" );
      return false;
   }
}

unsigned MU_Hide::size()
{
   return sizeof(MU_Hide) + m_diff.size() * sizeof(HideDifference);
}

void MU_Hide::setHideDifference( int number, bool visible )
{
   HideDifferenceList::iterator it;

   // Change selection state if it exists in our list
   for ( it = m_diff.begin(); it != m_diff.end(); it++ )
   {
      if ( (*it).number == number )
      {
         (*it).visible = visible;
         return;
      }
   }

   // add selection state to our list
   HideDifference diff;
   diff.number  = number;
   diff.visible = visible;

   m_diff.push_back( diff );
}

MU_InvertNormal::MU_InvertNormal()
{
}

MU_InvertNormal::~MU_InvertNormal()
{
}

void MU_InvertNormal::undo( Model * model )
{
   log_debug( "undo invert normal\n" );
   list<int>::iterator it;
   for ( it = m_triangles.begin(); it != m_triangles.end(); it++ )
   {
      model->invertNormals( *it );
   }
}

void MU_InvertNormal::redo( Model * model )
{
   list<int>::reverse_iterator it;
   for ( it = m_triangles.rbegin(); it != m_triangles.rend(); it++ )
   {
      model->invertNormals( *it );
   }
}

bool MU_InvertNormal::combine( Undo * u )
{
   MU_InvertNormal * undo = dynamic_cast< MU_InvertNormal * >( u );

   if ( undo )
   {
      list<int>::iterator it;
      for ( it = undo->m_triangles.begin(); it != undo->m_triangles.end(); it++ )
      {
         addTriangle( *it );
      }

      return true;
   }
   else
   {
      log_debug( "couldn't combine with invert normal\n" );
      return false;
   }
}

unsigned MU_InvertNormal::size()
{
   return sizeof(MU_InvertNormal) + m_triangles.size() * sizeof(int);
}

void MU_InvertNormal::addTriangle( int triangle )
{
   list<int>::iterator it;

   for ( it = m_triangles.begin(); it != m_triangles.end(); it++ )
   {
      if ( triangle == *it )
      {
         m_triangles.remove( triangle );
         return;
      }
   }

   m_triangles.push_back( triangle );
}

MU_MoveVertex::MU_MoveVertex()
{
}

MU_MoveVertex::~MU_MoveVertex()
{
}

void MU_MoveVertex::undo( Model * model )
{
   log_debug( "undo move vertex\n" );

   MoveVertexList::iterator it;

   // Modify a vertex we already have
   for ( it = m_vertices.begin(); it != m_vertices.end(); it++ )
   {
      if ( !(*it).isJoint )
      {
         model->moveVertex( (*it).number, (*it).oldx, (*it).oldy, (*it).oldz );
      }
      else
      {
         model->moveBoneJoint( (*it).number, (*it).oldx, (*it).oldy, (*it).oldz );
      }
   }
}

void MU_MoveVertex::redo( Model * model )
{
   MoveVertexList::iterator it;

   // Modify a vertex we already have
   for ( it = m_vertices.begin(); it != m_vertices.end(); it++ )
   {
      if ( !(*it).isJoint )
      {
         model->moveVertex( (*it).number, (*it).x, (*it).y, (*it).z );
      }
      else
      {
         model->moveBoneJoint( (*it).number, (*it).x, (*it).y, (*it).z );
      }
   }
}

bool MU_MoveVertex::combine( Undo * u )
{
   MU_MoveVertex * undo = dynamic_cast< MU_MoveVertex * >( u );

   if ( undo )
   {
      MoveVertexList::iterator it;

      for ( it = undo->m_vertices.begin(); it != undo->m_vertices.end(); it++ )
      {
         if ( !(*it).isJoint )
         {
            addVertex( (*it).number, (*it).x, (*it).y, (*it).z,
                  (*it).oldx, (*it).oldy, (*it).oldz );
         }
         else
         {
            addBoneJoint( (*it).number, (*it).x, (*it).y, (*it).z,
                  (*it).oldx, (*it).oldy, (*it).oldz );
         }
      }

      return true;
   }
   else
   {
      log_debug( "couldn't combine with move vertex\n" );
      return false;
   }
}

unsigned MU_MoveVertex::size()
{
   return sizeof(MU_MoveVertex) + m_vertices.size() * sizeof(MoveVertex);
}

void MU_MoveVertex::addVertex( int v, double x, double y, double z,
      double oldx, double oldy, double oldz )
{
   unsigned index;
   MoveVertex mv;
   mv.number = v;
   mv.isJoint = false;

   // Modify a vertex we already have
   if ( m_vertices.find_sorted( mv, index ) )
   {
      m_vertices[index].x = x;
      m_vertices[index].y = y;
      m_vertices[index].z = z;
      return;
   }

   // Not found, add new vertex information
   mv.x       = x;
   mv.y       = y;
   mv.z       = z;
   mv.oldx    = oldx;
   mv.oldy    = oldy;
   mv.oldz    = oldz;
   m_vertices.insert_sorted( mv );
}

void MU_MoveVertex::addBoneJoint( int v, double x, double y, double z,
      double oldx, double oldy, double oldz )
{
   unsigned index;
   MoveVertex mv;
   mv.number = v;
   mv.isJoint = true;

   // Modify a joint we already have
   if ( m_vertices.find_sorted( mv, index ) )
   {
      m_vertices[index].x = x;
      m_vertices[index].y = y;
      m_vertices[index].z = z;
      return;
   }

   // Not found, add new joint information
   mv.x       = x;
   mv.y       = y;
   mv.z       = z;
   mv.oldx    = oldx;
   mv.oldy    = oldy;
   mv.oldz    = oldz;
   m_vertices.insert_sorted( mv );
}

MU_SetTexture::MU_SetTexture()
{
}

MU_SetTexture::~MU_SetTexture()
{
}

void MU_SetTexture::undo( Model * model )
{
   log_debug( "undo set texture\n" );
   SetTextureList::iterator it;

   for ( it = m_list.begin(); it != m_list.end(); it++ )
   {
      model->setGroupTextureId( (*it).groupNumber, (*it).oldTexture );
   }
}

void MU_SetTexture::redo( Model * model )
{
   SetTextureList::iterator it;

   for ( it = m_list.begin(); it != m_list.end(); it++ )
   {
      model->setGroupTextureId( (*it).groupNumber, (*it).newTexture );
   }
}

bool MU_SetTexture::combine( Undo * u )
{
   MU_SetTexture * undo = dynamic_cast< MU_SetTexture * >( u );

   if ( undo )
   {
      SetTextureList::iterator it;

      for ( it = undo->m_list.begin(); it != undo->m_list.end(); it++ )
      {
         setTexture( (*it).groupNumber, (*it).newTexture, (*it).oldTexture );
      }

      return true;
   }
   else
   {
      log_debug( "couldn't combine with set texture\n" );
      return false;
   }
}

unsigned MU_SetTexture::size()
{
   return sizeof(MU_SetTexture) + m_list.size() * sizeof(SetTexture);
}

void MU_SetTexture::setTexture( unsigned groupNumber, int newTexture, int oldTexture )
{
   SetTextureList::iterator it;

   for ( it = m_list.begin(); it != m_list.end(); it++ )
   {
      if ( (*it).groupNumber == groupNumber )
      {
         (*it).newTexture = newTexture;
         return;
      }
   }

   SetTexture st;
   st.groupNumber = groupNumber;
   st.newTexture  = newTexture;
   st.oldTexture  = oldTexture;
   m_list.push_back( st );
}

MU_AddVertex::MU_AddVertex()
{
}

MU_AddVertex::~MU_AddVertex()
{
}

void MU_AddVertex::undo( Model * model )
{
   log_debug( "undo add vertex\n" );

   AddVertexList::reverse_iterator it;
   for ( it = m_list.rbegin(); it != m_list.rend(); it++ )
   {
      model->removeVertex( (*it).index );
   }
}

void MU_AddVertex::redo( Model * model )
{
   AddVertexList::iterator it;
   for ( it = m_list.begin(); it != m_list.end(); it++ )
   {
      model->insertVertex( (*it).index, (*it).vertex );
   }
}

bool MU_AddVertex::combine( Undo * u )
{
   MU_AddVertex * undo = dynamic_cast<MU_AddVertex *>( u );

   if ( undo )
   {
      AddVertexList::iterator it;
      for ( it = undo->m_list.begin(); it != undo->m_list.end(); it++ )
      {
         addVertex( (*it).index, (*it).vertex );
      }
      return true;
   }
   else
   {
      log_debug( "couldn't combine with add vertex\n" );
      return false;
   }
}

void MU_AddVertex::redoRelease()
{
   AddVertexList::iterator it;
   for ( it = m_list.begin(); it != m_list.end(); it++ )
   {
      (*it).vertex->release();
   }
}

unsigned MU_AddVertex::size()
{
   return sizeof(MU_AddVertex) + m_list.size() * (sizeof(AddVertex) + sizeof(Model::Vertex));
}

void MU_AddVertex::addVertex( unsigned index, Model::Vertex * vertex )
{
   if ( vertex )
   {
      AddVertex v;
      v.index  = index;
      v.vertex = vertex;
      m_list.push_back( v );
   }
}

MU_AddTriangle::MU_AddTriangle()
{
}

MU_AddTriangle::~MU_AddTriangle()
{
}

void MU_AddTriangle::undo( Model * model )
{
   log_debug( "undo add triangle\n" );

   AddTriangleList::reverse_iterator it;
   for ( it = m_list.rbegin(); it != m_list.rend(); it++ )
   {
      model->removeTriangle( (*it).index );
   }
}

void MU_AddTriangle::redo( Model * model )
{
   AddTriangleList::iterator it;
   for ( it = m_list.begin(); it != m_list.end(); it++ )
   {
      model->insertTriangle( (*it).index, (*it).triangle );
   }
}

bool MU_AddTriangle::combine( Undo * u )
{
   MU_AddTriangle * undo = dynamic_cast<MU_AddTriangle *>( u );

   if ( undo )
   {
      AddTriangleList::iterator it;
      for ( it = undo->m_list.begin(); it != undo->m_list.end(); it++ )
      {
         addTriangle( (*it).index, (*it).triangle );
      }
      return true;
   }
   else
   {
      log_debug( "couldn't combine with add triangle\n" );
      return false;
   }
}

void MU_AddTriangle::redoRelease()
{
   AddTriangleList::iterator it;
   for ( it = m_list.begin(); it != m_list.end(); it++ )
   {
      (*it).triangle->release();
   }
}

unsigned MU_AddTriangle::size()
{
   return sizeof(MU_AddTriangle) + m_list.size() * (sizeof(AddTriangle) + sizeof(Model::Triangle));
}

void MU_AddTriangle::addTriangle( unsigned index, Model::Triangle * triangle )
{
   if ( triangle )
   {
      AddTriangle v;
      v.index  = index;
      v.triangle = triangle;
      m_list.push_back( v );
   }
}

MU_AddGroup::MU_AddGroup()
{
}

MU_AddGroup::~MU_AddGroup()
{
}

void MU_AddGroup::undo( Model * model )
{
   log_debug( "undo add group\n" );
   AddGroupList::reverse_iterator it;
   for ( it = m_list.rbegin(); it != m_list.rend(); it++ )
   {
      model->removeGroup( (*it).index );
   }
}

void MU_AddGroup::redo( Model * model )
{
   AddGroupList::iterator it;
   for ( it = m_list.begin(); it != m_list.end(); it++ )
   {
      model->insertGroup( (*it).index, (*it).group );
   }
}

bool MU_AddGroup::combine( Undo * u )
{
   MU_AddGroup * undo = dynamic_cast<MU_AddGroup *>( u );

   if ( undo )
   {
      AddGroupList::iterator it;
      for ( it = undo->m_list.begin(); it != undo->m_list.end(); it++ )
      {
         addGroup( (*it).index, (*it).group );
      }
      return true;
   }
   else
   {
      log_debug( "couldn't combine with add group\n" );
      return false;
   }
}

void MU_AddGroup::redoRelease()
{
   AddGroupList::iterator it;
   for ( it = m_list.begin(); it != m_list.end(); it++ )
   {
      (*it).group->release();
   }
}

unsigned MU_AddGroup::size()
{
   return sizeof(MU_AddGroup) + m_list.size() * (sizeof(AddGroup) + sizeof(Model::Group));
}

void MU_AddGroup::addGroup( unsigned index, Model::Group * group )
{
   if ( group )
   {
      AddGroup v;
      v.index  = index;
      v.group  = group;
      m_list.push_back( v );
   }
}

MU_AddTexture::MU_AddTexture()
{
}

MU_AddTexture::~MU_AddTexture()
{
}

void MU_AddTexture::undo( Model * model )
{
   log_debug( "undo add texture\n" );
   AddTextureList::reverse_iterator it;
   for ( it = m_list.rbegin(); it != m_list.rend(); it++ )
   {
      model->removeTexture( (*it).index );
   }
}

void MU_AddTexture::redo( Model * model )
{
   AddTextureList::iterator it;
   for ( it = m_list.begin(); it != m_list.end(); it++ )
   {
      model->insertTexture( (*it).index, (*it).texture );
   }
}

bool MU_AddTexture::combine( Undo * u )
{
   MU_AddTexture * undo = dynamic_cast<MU_AddTexture *>( u );

   if ( undo )
   {
      AddTextureList::iterator it;
      for ( it = undo->m_list.begin(); it != undo->m_list.end(); it++ )
      {
         addTexture( (*it).index, (*it).texture );
      }
      return true;
   }
   else
   {
      log_debug( "couldn't combine with add texture\n" );
      return false;
   }
}

void MU_AddTexture::redoRelease()
{
   AddTextureList::iterator it;
   for ( it = m_list.begin(); it != m_list.end(); it++ )
   {
      (*it).texture->release();
   }
}

unsigned MU_AddTexture::size()
{
   return sizeof(MU_AddTexture) + m_list.size() * (sizeof(AddTexture) + sizeof(Model::Material));
}

void MU_AddTexture::addTexture( unsigned index, Model::Material * texture )
{
   if ( texture )
   {
      AddTexture v;
      v.index  = index;
      v.texture  = texture;
      m_list.push_back( v );
   }
}

MU_SetTextureCoords::MU_SetTextureCoords()
{
   m_list.clear();
}

MU_SetTextureCoords::~MU_SetTextureCoords()
{
}

void MU_SetTextureCoords::undo( Model * model )
{
   log_debug( "undo set texture coords\n" );
   STCList::iterator it;
   for ( it = m_list.begin(); it != m_list.end(); it++ )
   {
      model->setTextureCoords( (*it).triangle, (*it).vertexIndex, (*it).oldS, (*it).oldT );
   }
}

void MU_SetTextureCoords::redo( Model * model )
{
   STCList::iterator it;
   for ( it = m_list.begin(); it != m_list.end(); it++ )
   {
      model->setTextureCoords( (*it).triangle, (*it).vertexIndex, (*it).s, (*it).t );
   }
}

bool MU_SetTextureCoords::combine( Undo * u )
{
   MU_SetTextureCoords * undo = dynamic_cast<MU_SetTextureCoords *>( u );

   if ( undo )
   {
      STCList::iterator it;
      for ( it = undo->m_list.begin(); it != undo->m_list.end(); it++ )
      {
         addTextureCoords( (*it).triangle, (*it).vertexIndex, 
               (*it).s, (*it).t, (*it).oldS, (*it).oldT );
      }
      return true;
   }
   else
   {
      log_debug( "couldn't combine with set texture coords\n" );
      return false;
   }
}

unsigned MU_SetTextureCoords::size()
{
   return sizeof(MU_SetTextureCoords) + m_list.size() * sizeof(SetTextureCoords);
}

void MU_SetTextureCoords::addTextureCoords( unsigned triangle, unsigned vertexIndex, float s, float t, float oldS, float oldT )
{
   unsigned index;
   SetTextureCoords stc;
   stc.triangle    = triangle;
   stc.vertexIndex = vertexIndex;

   if ( m_list.find_sorted( stc, index ) )
   {
      m_list[index].s = s;
      m_list[index].t = t;
      return;
   }

   stc.s           = s;
   stc.t           = t;
   stc.oldS        = oldS;
   stc.oldT        = oldT;

   m_list.insert_sorted( stc );
}

MU_AddToGroup::MU_AddToGroup()
{
}

MU_AddToGroup::~MU_AddToGroup()
{
}

void MU_AddToGroup::undo( Model * model )
{
   log_debug( "undo add to group\n" );
   AddToGroupList::iterator it;
   for ( it = m_list.begin(); it != m_list.end(); it++ )
   {
      model->removeTriangleFromGroup( (*it).groupNum, (*it).triangleNum );
   }
}

void MU_AddToGroup::redo( Model * model )
{
   AddToGroupList::iterator it;
   for ( it = m_list.begin(); it != m_list.end(); it++ )
   {
      model->addTriangleToGroup( (*it).groupNum, (*it).triangleNum );
   }
}

bool MU_AddToGroup::combine( Undo * u )
{
   MU_AddToGroup * undo = dynamic_cast<MU_AddToGroup *>( u );

   if ( undo )
   {
      AddToGroupList::iterator it;
      for ( it = undo->m_list.begin(); it != undo->m_list.end(); it++ )
      {
         addToGroup( (*it).groupNum, (*it).triangleNum );
      }
      return true;
   }
   else
   {
      log_debug( "couldn't combine with add to group\n" );
      return false;
   }
}

unsigned MU_AddToGroup::size()
{
   return sizeof(MU_AddToGroup) + m_list.size() * sizeof(AddToGroup);
}

void MU_AddToGroup::addToGroup( unsigned groupNum, unsigned triangleNum )
{
   AddToGroupList::iterator it;
   for ( it = m_list.begin(); it != m_list.end(); it++ )
   {
      if ( (*it).triangleNum == triangleNum )
      {
         (*it).groupNum = groupNum;
         return;
      }
   }

   AddToGroup add;
   add.groupNum    = groupNum;
   add.triangleNum = triangleNum;
   m_list.push_back( add );
}

MU_RemoveFromGroup::MU_RemoveFromGroup()
{
}

MU_RemoveFromGroup::~MU_RemoveFromGroup()
{
}

void MU_RemoveFromGroup::undo( Model * model )
{
   log_debug( "undo remove from group\n" );
   RemoveFromGroupList::iterator it;
   for ( it = m_list.begin(); it != m_list.end(); it++ )
   {
      model->addTriangleToGroup( (*it).groupNum, (*it).triangleNum );
   }
}

void MU_RemoveFromGroup::redo( Model * model )
{
   RemoveFromGroupList::iterator it;
   for ( it = m_list.begin(); it != m_list.end(); it++ )
   {
      model->removeTriangleFromGroup( (*it).groupNum, (*it).triangleNum );
   }
}

bool MU_RemoveFromGroup::combine( Undo * u )
{
   MU_RemoveFromGroup * undo = dynamic_cast<MU_RemoveFromGroup *>( u );

   if ( undo )
   {
      RemoveFromGroupList::iterator it;
      for ( it = undo->m_list.begin(); it != undo->m_list.end(); it++ )
      {
         removeFromGroup( (*it).groupNum, (*it).triangleNum );
      }
      return true;
   }
   else
   {
      log_debug( "couldn't combine with remove from group\n" );
      return false;
   }
}

unsigned MU_RemoveFromGroup::size()
{
   return sizeof(MU_RemoveFromGroup) + m_list.size() * sizeof(RemoveFromGroup);
}

void MU_RemoveFromGroup::removeFromGroup( unsigned groupNum, unsigned triangleNum )
{
   /*
   RemoveFromGroupList::iterator it;
   for ( it = m_list.begin(); it != m_list.end(); it++ )
   {
      if ( (*it).triangleNum == triangleNum )
      {
         (*it).groupNum = groupNum;
         return;
      }
   }
   */

   RemoveFromGroup add;
   add.groupNum    = groupNum;
   add.triangleNum = triangleNum;
   m_list.push_back( add );
}

MU_DeleteTriangle::MU_DeleteTriangle()
{
}

MU_DeleteTriangle::~MU_DeleteTriangle()
{
}

void MU_DeleteTriangle::undo( Model * model )
{
   log_debug( "undo delete triangle\n" );
   DeleteTriangleList::reverse_iterator it;

   for ( it = m_list.rbegin(); it != m_list.rend(); it++ )
   {
      model->insertTriangle( (*it).triangleNum, (*it).triangle );
   }
}

void MU_DeleteTriangle::redo( Model * model )
{
   DeleteTriangleList::iterator it;

   for ( it = m_list.begin(); it != m_list.end(); it++ )
   {
      model->removeTriangle( (*it).triangleNum );
   }
}

bool MU_DeleteTriangle::combine( Undo * u )
{
   MU_DeleteTriangle * undo = dynamic_cast<MU_DeleteTriangle *>( u );

   if ( undo )
   {
      DeleteTriangleList::iterator it;
      for ( it = undo->m_list.begin(); it != undo->m_list.end(); it++ )
      {
         deleteTriangle( (*it).triangleNum, (*it).triangle );
      }

      return true;
   }
   else
   {
      log_debug( "couldn't combine with delete triangle\n" );
      return false;
   }
}

void MU_DeleteTriangle::undoRelease()
{
   DeleteTriangleList::iterator it;
   for ( it = m_list.begin(); it != m_list.end(); it++ )
   {
      (*it).triangle->release();
   }
}

unsigned MU_DeleteTriangle::size()
{
   return sizeof(MU_DeleteTriangle) + m_list.size() * (sizeof(DeleteTriangle) + sizeof(Model::Triangle));
}

void MU_DeleteTriangle::deleteTriangle( unsigned triangleNum, Model::Triangle * triangle )
{
   DeleteTriangle del;

   del.triangleNum = triangleNum;
   del.triangle    = triangle;

   m_list.push_back( del );
}

MU_DeleteVertex::MU_DeleteVertex()
{
}

MU_DeleteVertex::~MU_DeleteVertex()
{
}

void MU_DeleteVertex::undo( Model * model )
{
   log_debug( "undo delete vertex\n" );
   DeleteVertexList::reverse_iterator it;

   for ( it = m_list.rbegin(); it != m_list.rend(); it++ )
   {
      model->insertVertex( (*it).vertexNum, (*it).vertex );
   }
}

void MU_DeleteVertex::redo( Model * model )
{
   DeleteVertexList::iterator it;

   for ( it = m_list.begin(); it != m_list.end(); it++ )
   {
      model->removeVertex( (*it).vertexNum );
   }
}

bool MU_DeleteVertex::combine( Undo * u )
{
   MU_DeleteVertex * undo = dynamic_cast<MU_DeleteVertex *>( u );

   if ( undo )
   {
      DeleteVertexList::iterator it;
      for ( it = undo->m_list.begin(); it != undo->m_list.end(); it++ )
      {
         deleteVertex( (*it).vertexNum, (*it).vertex );
      }

      return true;
   }
   else
   {
      log_debug( "couldn't combine with delete vertex\n" );
      return false;
   }
}

void MU_DeleteVertex::undoRelease()
{
   DeleteVertexList::iterator it;
   for ( it = m_list.begin(); it != m_list.end(); it++ )
   {
      (*it).vertex->release();
   }
}

unsigned MU_DeleteVertex::size()
{
   return sizeof(MU_DeleteVertex) + m_list.size() * (sizeof(DeleteVertex) + sizeof(Model::Vertex));
}

void MU_DeleteVertex::deleteVertex( unsigned vertexNum, Model::Vertex * vertex )
{
   DeleteVertex del;

   del.vertexNum = vertexNum;
   del.vertex    = vertex;

   m_list.push_back( del );
}

MU_DeleteGroup::MU_DeleteGroup()
{
}

MU_DeleteGroup::~MU_DeleteGroup()
{
}

void MU_DeleteGroup::undo( Model * model )
{
   log_debug( "undo delete group\n" );
   DeleteGroupList::reverse_iterator it;

   for ( it = m_list.rbegin(); it != m_list.rend(); it++ )
   {
      model->insertGroup( (*it).groupNum, (*it).group );
   }
}

void MU_DeleteGroup::redo( Model * model )
{
   DeleteGroupList::iterator it;

   for ( it = m_list.begin(); it != m_list.end(); it++ )
   {
      model->removeGroup( (*it).groupNum );
   }
}

bool MU_DeleteGroup::combine( Undo * u )
{
   MU_DeleteGroup * undo = dynamic_cast<MU_DeleteGroup *>( u );

   if ( undo )
   {
      DeleteGroupList::iterator it;
      for ( it = undo->m_list.begin(); it != undo->m_list.end(); it++ )
      {
         deleteGroup( (*it).groupNum, (*it).group );
      }

      return true;
   }
   else
   {
      log_debug( "couldn't combine with delete group\n" );
      return false;
   }
}

void MU_DeleteGroup::undoRelease()
{
   DeleteGroupList::iterator it;
   for ( it = m_list.begin(); it != m_list.end(); it++ )
   {
      (*it).group->release();
   }
}

unsigned MU_DeleteGroup::size()
{
   return sizeof(MU_DeleteGroup) + m_list.size() * (sizeof(DeleteGroup) + sizeof(Model::Group));
}

void MU_DeleteGroup::deleteGroup( unsigned groupNum, Model::Group * group )
{
   DeleteGroup del;

   del.groupNum = groupNum;
   del.group    = group;

   m_list.push_back( del );
}

MU_DeleteTexture::MU_DeleteTexture()
{
}

MU_DeleteTexture::~MU_DeleteTexture()
{
}

void MU_DeleteTexture::undo( Model * model )
{
   log_debug( "undo delete texture\n" );
   DeleteTextureList::reverse_iterator it;

   for ( it = m_list.rbegin(); it != m_list.rend(); it++ )
   {
      model->insertTexture( (*it).textureNum, (*it).texture );
   }
}

void MU_DeleteTexture::redo( Model * model )
{
   DeleteTextureList::iterator it;

   for ( it = m_list.begin(); it != m_list.end(); it++ )
   {
      model->removeTexture( (*it).textureNum );
   }
}

bool MU_DeleteTexture::combine( Undo * u )
{
   MU_DeleteTexture * undo = dynamic_cast<MU_DeleteTexture *>( u );

   if ( undo )
   {
      DeleteTextureList::iterator it;
      for ( it = undo->m_list.begin(); it != undo->m_list.end(); it++ )
      {
         deleteTexture( (*it).textureNum, (*it).texture );
      }

      return true;
   }
   else
   {
      log_debug( "couldn't combine with delete texture\n" );
      return false;
   }
}

void MU_DeleteTexture::undoRelease()
{
   DeleteTextureList::iterator it;
   for ( it = m_list.begin(); it != m_list.end(); it++ )
   {
      (*it).texture->release();
   }
}

unsigned MU_DeleteTexture::size()
{
   return sizeof(MU_DeleteTexture) + m_list.size() * (sizeof(DeleteTexture) + sizeof(Model::Material));
}

void MU_DeleteTexture::deleteTexture( unsigned textureNum, Model::Material * texture )
{
   DeleteTexture del;

   del.textureNum = textureNum;
   del.texture    = texture;

   m_list.push_back( del );
}

MU_SetLightProperties::MU_SetLightProperties()
{
}

MU_SetLightProperties::~MU_SetLightProperties()
{
}

void MU_SetLightProperties::undo( Model * model )
{
   log_debug( "undo set light properties\n" );
   LightPropertiesList::reverse_iterator it;

   for ( it = m_list.rbegin(); it != m_list.rend(); it++ )
   {
      if ( (*it).isSet[0] )
      {
         model->setTextureAmbient( (*it).textureNum, (*it).oldLight[0] );
      }
      if ( (*it).isSet[1] )
      {
         model->setTextureDiffuse( (*it).textureNum, (*it).oldLight[1] );
      }
      if ( (*it).isSet[2] )
      {
         model->setTextureSpecular( (*it).textureNum, (*it).oldLight[2] );
      }
      if ( (*it).isSet[3] )
      {
         model->setTextureEmissive( (*it).textureNum, (*it).oldLight[3] );
      }
   }
}

void MU_SetLightProperties::redo( Model * model )
{
   LightPropertiesList::iterator it;

   for ( it = m_list.begin(); it != m_list.end(); it++ )
   {
      if ( (*it).isSet[0] )
      {
         model->setTextureAmbient( (*it).textureNum, (*it).newLight[0] );
      }
      if ( (*it).isSet[1] )
      {
         model->setTextureDiffuse( (*it).textureNum, (*it).newLight[1] );
      }
      if ( (*it).isSet[2] )
      {
         model->setTextureSpecular( (*it).textureNum, (*it).newLight[2] );
      }
      if ( (*it).isSet[3] )
      {
         model->setTextureEmissive( (*it).textureNum, (*it).newLight[3] );
      }
   }
}

bool MU_SetLightProperties::combine( Undo * u )
{
   MU_SetLightProperties * undo = dynamic_cast<MU_SetLightProperties *>( u );

   if ( undo )
   {
      LightPropertiesList::iterator it;
      for ( it = undo->m_list.begin(); it != undo->m_list.end(); it++ )
      {
         for ( int t = 0; t < LightTypeMax; t++ )
         {
            if ( (*it).isSet[t] )
            {
               setLightProperties( (*it).textureNum, (LightType) t, (*it).newLight[t], (*it).oldLight[t] );
            }
         }
      }

      return true;
   }
   else
   {
      log_debug( "couldn't combine with set light properties\n" );
      return false;
   }
}

unsigned MU_SetLightProperties::size()
{
   return sizeof(MU_SetLightProperties) + m_list.size() * sizeof(LightProperties);
}

void MU_SetLightProperties::setLightProperties( unsigned textureNum, LightType type, const float * newLight, const float * oldLight )
{
   LightPropertiesList::iterator it;

   for ( it = m_list.begin(); it != m_list.end(); it++ )
   {
      if ( (*it).textureNum == textureNum )
      {
         // Set old light if this is the first time we set this type
         if ( ! (*it).isSet[type] )
         {
            for ( int n = 0; n < 4; n++ )
            {
               (*it).oldLight[type][n] = oldLight[n];
            }
         }

         // Set new light for this type
         (*it).isSet[ type ] = true;
         log_error( "light %d:\n", type  );
         for ( int n = 0; n < 4; n++ )
         {
            log_error( "  %f %f\n", oldLight[n], newLight[n] );
            (*it).newLight[type][n] = newLight[n];
         }
         return;
      }
   }

   // Add new LightProperties to list
   LightProperties prop;

   for ( int t = 0; t < LightTypeMax; t++ )
   {
      prop.isSet[t] = false;
   }

   prop.textureNum = textureNum;
   prop.isSet[ type ] = true;
   log_error( "light %d:\n", type  );
   for ( int n = 0; n < 4; n++ )
   {
      log_error( "  %f %f\n", oldLight[n], newLight[n] );
      prop.oldLight[type][n] = oldLight[n];
      prop.newLight[type][n] = newLight[n];
   }

   m_list.push_back( prop );
}

MU_SetShininess::MU_SetShininess()
{
}

MU_SetShininess::~MU_SetShininess()
{
}

void MU_SetShininess::undo( Model * model )
{
   log_debug( "undo set shininess\n" );
   ShininessList::reverse_iterator it;

   for ( it = m_list.rbegin(); it != m_list.rend(); it++ )
   {
      model->setTextureShininess( (*it).textureNum, (*it).oldValue );
   }
}

void MU_SetShininess::redo( Model * model )
{
   ShininessList::iterator it;

   for ( it = m_list.begin(); it != m_list.end(); it++ )
   {
      model->setTextureShininess( (*it).textureNum, (*it).newValue );
   }
}

bool MU_SetShininess::combine( Undo * u )
{
   MU_SetShininess * undo = dynamic_cast<MU_SetShininess *>( u );

   if ( undo )
   {
      ShininessList::iterator it;
      for ( it = undo->m_list.begin(); it != undo->m_list.end(); it++ )
      {
         setShininess( (*it).textureNum, (*it).newValue, (*it).oldValue );
      }

      return true;
   }
   else
   {
      log_debug( "couldn't combine with set shininess\n" );
      return false;
   }
}

unsigned MU_SetShininess::size()
{
   return sizeof(MU_SetShininess) + m_list.size() * sizeof(Shininess);
}

void MU_SetShininess::setShininess( unsigned textureNum, const float & newValue, const float & oldValue )
{
   ShininessList::iterator it;

   for ( it = m_list.begin(); it != m_list.end(); it++ )
   {
      if ( (*it).textureNum == textureNum )
      {
         (*it).newValue = newValue;
         return;
      }
   }

   // Add new Shininess to list
   Shininess shine;

   shine.textureNum = textureNum;
   shine.oldValue = oldValue;
   shine.newValue = newValue;

   m_list.push_back( shine );
}

MU_SetTextureName::MU_SetTextureName()
{
}

MU_SetTextureName::~MU_SetTextureName()
{
}

void MU_SetTextureName::undo( Model * model )
{
   log_debug( "undo set texture name\n" );
   model->setTextureName( m_textureNum, m_oldName.c_str() );
}

void MU_SetTextureName::redo( Model * model )
{
   model->setTextureName( m_textureNum, m_newName.c_str() );
}

bool MU_SetTextureName::combine( Undo * u )
{
   MU_SetTextureName * undo = dynamic_cast< MU_SetTextureName * >( u );

   if ( undo && undo->m_textureNum == m_textureNum )
   {
      m_newName = undo->m_newName;
      return true;
   }
   else
   {
      return false;
   }
}

unsigned MU_SetTextureName::size()
{
   return sizeof(MU_SetTextureName);
}

void MU_SetTextureName::setTextureName( unsigned textureNum, const char * newName, const char * oldName )
{
   m_textureNum = textureNum;
   m_newName    = newName;
   m_oldName    = oldName;
}

MU_SetTriangleVertices::MU_SetTriangleVertices()
{
}

MU_SetTriangleVertices::~MU_SetTriangleVertices()
{
}

void MU_SetTriangleVertices::undo( Model * model )
{
   log_debug( "undo set triangle vertices\n" );
   TriangleVerticesList::reverse_iterator it;

   for ( it = m_list.rbegin(); it != m_list.rend(); it++ )
   {
      model->setTriangleVertices( (*it).triangleNum, 
            (*it).oldVertices[0],
            (*it).oldVertices[1],
            (*it).oldVertices[2] );
   }
}

void MU_SetTriangleVertices::redo( Model * model )
{
   TriangleVerticesList::iterator it;

   for ( it = m_list.begin(); it != m_list.end(); it++ )
   {
      model->setTriangleVertices( (*it).triangleNum, 
            (*it).newVertices[0],
            (*it).newVertices[1],
            (*it).newVertices[2] );
   }
}

bool MU_SetTriangleVertices::combine( Undo * u )
{
   MU_SetTriangleVertices * undo = dynamic_cast<MU_SetTriangleVertices *>( u );

   if ( undo )
   {
      TriangleVerticesList::iterator it;
      for ( it = undo->m_list.begin(); it != undo->m_list.end(); it++ )
      {
         setTriangleVertices( (*it).triangleNum, 
               (*it).newVertices[0], 
               (*it).newVertices[1], 
               (*it).newVertices[2], 
               (*it).oldVertices[0], 
               (*it).oldVertices[1], 
               (*it).oldVertices[2] );
      }

      return true;
   }
   else
   {
      log_debug( "couldn't combine with set triangle vertices\n" );
      return false;
   }
}

unsigned MU_SetTriangleVertices::size()
{
   return sizeof(MU_SetTriangleVertices) + m_list.size() * sizeof(TriangleVertices);
}

void MU_SetTriangleVertices::setTriangleVertices( unsigned triangleNum, 
      unsigned newV1, unsigned newV2, unsigned newV3, 
      unsigned oldV1, unsigned oldV2, unsigned oldV3 )
{
   TriangleVerticesList::iterator it;

   for ( it = m_list.begin(); it != m_list.end(); it++ )
   {
      if ( (*it).triangleNum == triangleNum )
      {
         (*it).newVertices[0] = newV1;
         (*it).newVertices[1] = newV2;
         (*it).newVertices[2] = newV3;
         return;
      }
   }

   // Add new Shininess to list
   TriangleVertices tv;

   tv.triangleNum = triangleNum;
   tv.oldVertices[0] = oldV1;
   tv.oldVertices[1] = oldV2;
   tv.oldVertices[2] = oldV3;
   tv.newVertices[0] = newV1;
   tv.newVertices[1] = newV2;
   tv.newVertices[2] = newV3;

   m_list.push_back( tv );
}

MU_SubdivideSelected::MU_SubdivideSelected()
{
}

MU_SubdivideSelected::~MU_SubdivideSelected()
{
}

void MU_SubdivideSelected::undo( Model * model )
{
   log_debug( "undo subdivide selected\n" );
}

void MU_SubdivideSelected::redo( Model * model )
{
   model->subdivideSelectedTriangles();
}

bool MU_SubdivideSelected::combine( Undo * u )
{
   log_debug( "couldn't combine with subdivide selected\n" );
   return false;
}

MU_SubdivideTriangle::MU_SubdivideTriangle()
{
}

MU_SubdivideTriangle::~MU_SubdivideTriangle()
{
}

void MU_SubdivideTriangle::undo( Model * model )
{
   log_debug( "undo subdivide\n" );

   SubdivideTriangleList::reverse_iterator it;

   for ( it = m_list.rbegin(); it != m_list.rend(); it++ )
   {
      model->unsubdivideTriangles( (*it).a, (*it).b, (*it).c, (*it).d );
   }

   list<unsigned>::reverse_iterator iit;

   for ( iit = m_vlist.rbegin(); iit != m_vlist.rend(); iit++ )
   {
      model->deleteVertex( *iit );
   }
}

void MU_SubdivideTriangle::redo( Model * model )
{
}

bool MU_SubdivideTriangle::combine( Undo * u )
{
   MU_SubdivideTriangle * undo = dynamic_cast<MU_SubdivideTriangle *>( u );
   if ( undo )
   {
      SubdivideTriangle st;

      while ( ! undo->m_list.empty() )
      {
         st = undo->m_list.front();
         subdivide( st.a, st.b, st.c, st.d );
         undo->m_list.pop_front();
      }

      while ( ! undo->m_vlist.empty() )
      {
         addVertex( undo->m_vlist.front() );
         undo->m_vlist.pop_front();
      }

      return true;
   }
   else
   {
      log_debug( "couldn't combine with subdivide triangles\n" );
      return false;
   }
}

unsigned MU_SubdivideTriangle::size()
{
   return sizeof( MU_SubdivideTriangle ) 
      + m_list.size()  * sizeof( SubdivideTriangle )
      + m_vlist.size() * sizeof(int);
}

void MU_SubdivideTriangle::subdivide( unsigned a, unsigned b, unsigned c, unsigned d )
{
   SubdivideTriangle st;

   st.a = a;
   st.b = b;
   st.c = c;
   st.d = d;

   m_list.push_back( st );
}

void MU_SubdivideTriangle::addVertex( unsigned v )
{
   m_vlist.push_back( v );
}

MU_ChangeAnimState::MU_ChangeAnimState()
{
}

MU_ChangeAnimState::~MU_ChangeAnimState()
{
}

void MU_ChangeAnimState::undo( Model * model )
{
   log_debug( "undo change anim state: old %d\n", m_oldMode );
   if ( m_oldMode )
   {
      model->setCurrentAnimation( m_oldMode, m_anim );
      model->setCurrentAnimationFrame( m_frame );
   }
   else
   {
      model->setNoAnimation();
   }
}

void MU_ChangeAnimState::redo( Model * model )
{
   log_debug( "redo change anim state: new %d\n", m_newMode );
   if ( m_newMode )
   {
      model->setCurrentAnimation( m_newMode, m_anim );
      model->setCurrentAnimationFrame( m_frame );
   }
   else
   {
      model->setNoAnimation();
   }
}

bool MU_ChangeAnimState::combine( Undo * u )
{
   return false;
}

unsigned MU_ChangeAnimState::size()
{
   return sizeof( MU_ChangeAnimState );
}

void MU_ChangeAnimState::setState( Model::AnimationMode newMode, Model::AnimationMode oldMode, unsigned anim, unsigned frame )
{
   log_debug( "ChangeAnimState undo info: old = %d, new = %d\n", oldMode, newMode );
   m_newMode = newMode;
   m_oldMode = oldMode;
   m_anim    = anim;
   m_frame   = frame;
}

MU_SetAnimName::MU_SetAnimName()
{
}

MU_SetAnimName::~MU_SetAnimName()
{
}

void MU_SetAnimName::undo( Model * model )
{
   model->setAnimName( m_mode, m_animNum, m_oldName.c_str() );
   if ( model->getAnimationMode() != Model::ANIMMODE_NONE && model->getAnimationMode() != m_mode || model->getCurrentAnimation() != m_animNum )
   {
      model->setCurrentAnimation( m_mode, m_oldName.c_str() );
   }
}

void MU_SetAnimName::redo( Model * model )
{
   model->setAnimName( m_mode, m_animNum, m_newName.c_str() );
   if ( model->getAnimationMode() != Model::ANIMMODE_NONE && model->getAnimationMode() != m_mode || model->getCurrentAnimation() != m_animNum )
   {
      model->setCurrentAnimation( m_mode, m_newName.c_str() );
   }
}

bool MU_SetAnimName::combine( Undo * u )
{
   return false;
}

unsigned MU_SetAnimName::size()
{
   return sizeof(MU_SetAnimName);
}

void MU_SetAnimName::setName( Model::AnimationMode mode, unsigned animNum, const char * newName, const char * oldName )
{
   m_mode       = mode;
   m_animNum    = animNum;
   m_newName    = newName;
   m_oldName    = oldName;
}

MU_SetAnimFrameCount::MU_SetAnimFrameCount()
{
}

MU_SetAnimFrameCount::~MU_SetAnimFrameCount()
{
}

void MU_SetAnimFrameCount::undo( Model * model )
{
   model->setAnimFrameCount( m_mode, m_animNum, m_oldCount );
}

void MU_SetAnimFrameCount::redo( Model * model )
{
   model->setAnimFrameCount( m_mode, m_animNum, m_newCount );
}

bool MU_SetAnimFrameCount::combine( Undo * u )
{
   return false;
}

unsigned MU_SetAnimFrameCount::size()
{
   return sizeof(MU_SetAnimFrameCount);
}

void MU_SetAnimFrameCount::setAnimFrameCount( Model::AnimationMode mode, unsigned animNum, unsigned newCount, unsigned oldCount )
{
   m_mode       = mode;
   m_animNum    = animNum;
   m_newCount   = newCount;
   m_oldCount   = oldCount;
}

MU_SetAnimFPS::MU_SetAnimFPS()
{
}

MU_SetAnimFPS::~MU_SetAnimFPS()
{
}

void MU_SetAnimFPS::undo( Model * model )
{
   model->setAnimFPS( m_mode, m_animNum, m_oldFPS );
   if ( model->getAnimationMode() != Model::ANIMMODE_NONE && (model->getAnimationMode() != m_mode || model->getCurrentAnimation() != m_animNum) ) 
   {
      model->setCurrentAnimation( m_mode, m_animNum );
   }
}

void MU_SetAnimFPS::redo( Model * model )
{
   model->setAnimFPS( m_mode, m_animNum, m_newFPS );
   if ( model->getAnimationMode() != Model::ANIMMODE_NONE && (model->getAnimationMode() != m_mode || model->getCurrentAnimation() != m_animNum) ) 
   {
      model->setCurrentAnimation( m_mode, m_animNum );
   }
}

bool MU_SetAnimFPS::combine( Undo * u )
{
   MU_SetAnimFPS * undo = dynamic_cast< MU_SetAnimFPS * >( u );

   if ( undo && undo->m_mode == m_mode && undo->m_animNum == m_animNum )
   {
      m_newFPS = undo->m_newFPS;
      return true;
   }
   else
   {
      log_debug( "couldn't combine with set anim fps\n" );
      return false;
   }
}

unsigned MU_SetAnimFPS::size()
{
   return sizeof(MU_SetAnimFPS);
}

void MU_SetAnimFPS::setFPS( Model::AnimationMode mode, unsigned animNum, double newFps, double oldFps )
{
   m_mode       = mode;
   m_animNum    = animNum;
   m_newFPS     = newFps;
   m_oldFPS     = oldFps;
}

MU_SetAnimKeyframe::MU_SetAnimKeyframe()
{
}

MU_SetAnimKeyframe::~MU_SetAnimKeyframe()
{
}

void MU_SetAnimKeyframe::undo( Model * model )
{
   SetKeyframeList::iterator it;

   for ( it = m_keyframes.begin(); it != m_keyframes.end(); it++ )
   {
      if ( (*it).isNew )
      {
         log_debug( "undoing new keyframe\n" );
         model->removeSkelAnimKeyframe( m_anim, m_frame, (*it).number, m_isRotation, true );
      }
      else
      {
         log_debug( "undoing existing keyframe\n" );
         model->setSkelAnimKeyframe( m_anim, m_frame, (*it).number, m_isRotation,
               (*it).oldx, (*it).oldy, (*it).oldz );
      }
   }
   if ( model->getAnimationMode() != Model::ANIMMODE_NONE ) 
   {
      model->setCurrentAnimation( Model::ANIMMODE_SKELETAL, m_anim );
      model->setCurrentAnimationFrame( m_frame );
   }
}

void MU_SetAnimKeyframe::redo( Model * model )
{
   SetKeyframeList::iterator it;

   for ( it = m_keyframes.begin(); it != m_keyframes.end(); it++ )
   {
      model->setSkelAnimKeyframe( m_anim, m_frame, (*it).number, m_isRotation,
            (*it).x, (*it).y, (*it).z );
   }

   if ( model->getAnimationMode() != Model::ANIMMODE_NONE ) 
   {
      model->setCurrentAnimation( Model::ANIMMODE_SKELETAL, m_anim );
      model->setCurrentAnimationFrame( m_frame );
   }
}

bool MU_SetAnimKeyframe::combine( Undo * u )
{
   MU_SetAnimKeyframe * undo = dynamic_cast< MU_SetAnimKeyframe * >( u );

   if ( undo && undo->m_anim == m_anim && undo->m_frame == m_frame && m_isRotation == undo->m_isRotation )
   {
      SetKeyframeList::iterator it;

      for ( it = undo->m_keyframes.begin(); it != undo->m_keyframes.end(); it++ )
      {
         addBoneJoint( (*it).number, (*it).isNew, (*it).x, (*it).y, (*it).z,
               (*it).oldx, (*it).oldy, (*it).oldz );
      }

      return true;
   }
   else
   {
      log_debug( "couldn't combine with set anim keyframe\n" );
      return false;
   }
}

unsigned MU_SetAnimKeyframe::size()
{
   return sizeof(MU_SetAnimKeyframe) + m_keyframes.size() * sizeof(SetKeyframe);
}

void MU_SetAnimKeyframe::setAnimationData( const unsigned & anim, const unsigned & frame, const bool & isRotation )
{
   m_anim       = anim;
   m_frame      = frame;
   m_isRotation = isRotation;
}

void MU_SetAnimKeyframe::addBoneJoint( int j, bool isNew, double x, double y, double z,
      double oldx, double oldy, double oldz )
{
   unsigned index = 0;
   SetKeyframe mv;
   mv.number = j;

   // Modify a joint we already have
   if ( m_keyframes.find_sorted( mv, index ) )
   {
      m_keyframes[index].x = x;
      m_keyframes[index].y = y;
      m_keyframes[index].z = z;
      return;
   }

   // Not found, add new joint information
   mv.isNew   = isNew;
   mv.x       = x;
   mv.y       = y;
   mv.z       = z;
   mv.oldx    = oldx;
   mv.oldy    = oldy;
   mv.oldz    = oldz;
   m_keyframes.insert_sorted( mv );
}

MU_DeleteKeyframe::MU_DeleteKeyframe()
{
}

MU_DeleteKeyframe::~MU_DeleteKeyframe()
{
}

void MU_DeleteKeyframe::undo( Model * model )
{
   log_debug( "undo delete keyframe\n" );
   DeleteKeyframeList::reverse_iterator it;

   unsigned frame = 0;
   for ( it = m_list.rbegin(); it != m_list.rend(); it++ )
   {
      model->insertSkelAnimKeyframe( m_anim, *it );
      frame = (*it)->m_frame;
   }

   if ( model->getAnimationMode() != Model::ANIMMODE_NONE )
   {
      if ( m_anim != model->getCurrentAnimation() )
      {
         model->setCurrentAnimation( Model::ANIMMODE_SKELETAL, m_anim );
      }
      model->setCurrentAnimationFrame( frame );
   }
}

void MU_DeleteKeyframe::redo( Model * model )
{
   DeleteKeyframeList::iterator it;

   unsigned frame = 0;
   for ( it = m_list.begin(); it != m_list.end(); it++ )
   {
      model->removeSkelAnimKeyframe( m_anim, (*it)->m_frame, (*it)->m_jointIndex, (*it)->m_isRotation );
      frame = (*it)->m_frame;
   }

   if ( model->getAnimationMode() != Model::ANIMMODE_NONE )
   {
      if ( m_anim != model->getCurrentAnimation() )
      {
         model->setCurrentAnimation( Model::ANIMMODE_SKELETAL, m_anim );
      }
      model->setCurrentAnimationFrame( frame );
   }
}

bool MU_DeleteKeyframe::combine( Undo * u )
{
   MU_DeleteKeyframe * undo = dynamic_cast<MU_DeleteKeyframe *>( u );

   if ( undo && m_anim == undo->m_anim )
   {
      DeleteKeyframeList::iterator it;
      for ( it = undo->m_list.begin(); it != undo->m_list.end(); it++ )
      {
         deleteKeyframe( *it );
      }

      return true;
   }
   else
   {
      log_debug( "couldn't combine with delete keyframe\n" );
      return false;
   }
}

void MU_DeleteKeyframe::undoRelease()
{
   DeleteKeyframeList::iterator it;
   for ( it = m_list.begin(); it != m_list.end(); it++ )
   {
      (*it)->release();
   }
}

unsigned MU_DeleteKeyframe::size()
{
   return sizeof(MU_DeleteKeyframe) + m_list.size() * sizeof(Model::Keyframe);
}

void MU_DeleteKeyframe::setAnimationData( const unsigned & anim )
{
   m_anim = anim;
}

void MU_DeleteKeyframe::deleteKeyframe( Model::Keyframe * keyframe )
{
   m_list.push_back( keyframe );
}

MU_SetJointName::MU_SetJointName()
{
}

MU_SetJointName::~MU_SetJointName()
{
}

void MU_SetJointName::undo( Model * model )
{
   model->setBoneJointName( m_joint, m_oldName.c_str() );
}

void MU_SetJointName::redo( Model * model )
{
   model->setBoneJointName( m_joint, m_newName.c_str() );
}

bool MU_SetJointName::combine( Undo * u )
{
   return false;
}

unsigned MU_SetJointName::size()
{
   return sizeof(MU_SetJointName);
}

void MU_SetJointName::setName( unsigned joint, const char * newName, const char * oldName )
{
   m_joint   = joint;
   m_newName = newName;
   m_oldName = oldName;
}

MU_MoveFrameVertex::MU_MoveFrameVertex()
{
}

MU_MoveFrameVertex::~MU_MoveFrameVertex()
{
}

void MU_MoveFrameVertex::undo( Model * model )
{
   MoveFrameVertexList::iterator it;

   // Modify a vertex we already have
   for ( it = m_vertices.begin(); it != m_vertices.end(); it++ )
   {
      model->setFrameAnimVertexCoords( m_anim, m_frame, (*it).number, (*it).oldx, (*it).oldy, (*it).oldz );
   }

   if ( model->getAnimationMode() != Model::ANIMMODE_NONE ) 
   {
      model->setCurrentAnimation( Model::ANIMMODE_FRAME, m_anim );
      model->setCurrentAnimationFrame( m_frame );
   }
}

void MU_MoveFrameVertex::redo( Model * model )
{
   MoveFrameVertexList::iterator it;

   // Modify a vertex we already have
   for ( it = m_vertices.begin(); it != m_vertices.end(); it++ )
   {
      model->setFrameAnimVertexCoords( m_anim, m_frame, (*it).number, (*it).x, (*it).y, (*it).z );
   }

   if ( model->getAnimationMode() != Model::ANIMMODE_NONE ) 
   {
      model->setCurrentAnimation( Model::ANIMMODE_FRAME, m_anim );
      model->setCurrentAnimationFrame( m_frame );
   }
}

bool MU_MoveFrameVertex::combine( Undo * u )
{
   MU_MoveFrameVertex * undo = dynamic_cast< MU_MoveFrameVertex * >( u );

   if ( undo )
   {
      MoveFrameVertexList::iterator it;

      for ( it = undo->m_vertices.begin(); it != undo->m_vertices.end(); it++ )
      {
         addVertex( (*it).number, (*it).x, (*it).y, (*it).z,
               (*it).oldx, (*it).oldy, (*it).oldz );
      }

      return true;
   }
   else
   {
      log_debug( "couldn't combine with move frame vertex\n" );
      return false;
   }
}

unsigned MU_MoveFrameVertex::size()
{
   return sizeof(MU_MoveFrameVertex) + m_vertices.size() * sizeof(MoveFrameVertex);
}

void MU_MoveFrameVertex::setAnimationData( const unsigned & anim, const unsigned & frame )
{
   m_anim = anim;
   m_frame = frame;
}

void MU_MoveFrameVertex::addVertex( int v, double x, double y, double z,
      double oldx, double oldy, double oldz )
{
   unsigned index;
   MoveFrameVertex mv;
   mv.number = v;

   // Modify a vertex we already have
   if ( m_vertices.find_sorted( mv, index ) )
   {
      m_vertices[index].x = x;
      m_vertices[index].y = y;
      m_vertices[index].z = z;
      return;
   }

   // Not found, add new vertex information
   mv.x       = x;
   mv.y       = y;
   mv.z       = z;
   mv.oldx    = oldx;
   mv.oldy    = oldy;
   mv.oldz    = oldz;
   m_vertices.insert_sorted( mv );
}

MU_AddFrameAnimFrame::MU_AddFrameAnimFrame()
{
}

MU_AddFrameAnimFrame::~MU_AddFrameAnimFrame()
{
}

void MU_AddFrameAnimFrame::undo( Model * model )
{
   log_debug( "undo add vertex\n" );

   FrameVertexList::reverse_iterator it;
   for ( it = m_list.rbegin(); it != m_list.rend(); it++ )
   {
      model->removeFrameAnimFrame( m_anim, (*it).frame );
   }

   if ( model->getAnimationMode() != Model::ANIMMODE_NONE ) 
   {
      model->setCurrentAnimation( Model::ANIMMODE_FRAME, m_anim );
      model->setCurrentAnimationFrame( 0 );
   }
}

void MU_AddFrameAnimFrame::redo( Model * model )
{
   FrameVertexList::iterator it;
   for ( it = m_list.begin(); it != m_list.end(); it++ )
   {
      model->insertFrameAnimFrame( m_anim, (*it).frame, (*it).list );
   }

   if ( model->getAnimationMode() != Model::ANIMMODE_NONE ) 
   {
      model->setCurrentAnimation( Model::ANIMMODE_FRAME, m_anim );
      model->setCurrentAnimationFrame( 0 );
   }
}

bool MU_AddFrameAnimFrame::combine( Undo * u )
{
   MU_AddFrameAnimFrame * undo = dynamic_cast<MU_AddFrameAnimFrame *>( u );

   if ( undo && m_anim == undo->m_anim )
   {
      FrameVertexList::iterator it;
      for ( it = undo->m_list.begin(); it != undo->m_list.end(); it++ )
      {
         addFrame( (*it).frame, (*it).list );
      }
      return true;
   }
   else
   {
      log_debug( "couldn't combine with add frame anim frame\n" );
      return false;
   }
}

void MU_AddFrameAnimFrame::redoRelease()
{
   FrameVertexList::iterator it;
   for ( it = m_list.begin(); it != m_list.end(); it++ )
   {
      for ( unsigned t = 0; t < (*it).list->size(); t++ )
      {
         (*(*it).list)[t]->release();
      }
      (*it).list->clear();
   }
}

unsigned MU_AddFrameAnimFrame::size()
{
   unsigned listSize = 0;
   FrameVertexList::iterator it;
   for ( it = m_list.begin(); it != m_list.end(); it++ )
   {
      listSize += sizeof( Model::FrameAnimVertexList ) + (*it).list->size() * sizeof( Model::FrameAnimVertex );
   }
   return sizeof(MU_AddFrameAnimFrame) + m_list.size() * sizeof(FrameVertexList) + listSize;
}

void MU_AddFrameAnimFrame::setAnimationData( const unsigned & anim )
{
   m_anim = anim;
}

void MU_AddFrameAnimFrame::addFrame( const unsigned & frame, Model::FrameAnimVertexList * list )
{
   if ( list )
   {
      AddFrame l;
      l.frame  = frame;
      l.list   = list;
      m_list.push_back( l );
   }
}

MU_DeleteFrameAnimFrame::MU_DeleteFrameAnimFrame()
{
}

MU_DeleteFrameAnimFrame::~MU_DeleteFrameAnimFrame()
{
}

void MU_DeleteFrameAnimFrame::undo( Model * model )
{
   log_debug( "undo delete frame anim frame\n" );
   FrameVertexList::reverse_iterator it;

   for ( it = m_list.rbegin(); it != m_list.rend(); it++ )
   {
      model->insertFrameAnimFrame( m_anim, (*it).frame, (*it).list );
   }

   if ( model->getAnimationMode() != Model::ANIMMODE_NONE ) 
   {
      model->setCurrentAnimation( Model::ANIMMODE_FRAME, m_anim );
      model->setCurrentAnimationFrame( 0 );
   }
}

void MU_DeleteFrameAnimFrame::redo( Model * model )
{
   FrameVertexList::iterator it;

   for ( it = m_list.begin(); it != m_list.end(); it++ )
   {
      model->removeFrameAnimFrame( m_anim, (*it).frame );
   }

   if ( model->getAnimationMode() != Model::ANIMMODE_NONE ) 
   {
      model->setCurrentAnimation( Model::ANIMMODE_FRAME, m_anim );
      model->setCurrentAnimationFrame( 0 );
   }
}

bool MU_DeleteFrameAnimFrame::combine( Undo * u )
{
   MU_DeleteFrameAnimFrame * undo = dynamic_cast<MU_DeleteFrameAnimFrame *>( u );

   if ( undo && m_anim == undo->m_anim )
   {
      FrameVertexList::iterator it;
      for ( it = undo->m_list.begin(); it != undo->m_list.end(); it++ )
      {
         deleteFrame( (*it).frame, (*it).list );
      }

      return true;
   }
   else
   {
      log_debug( "couldn't combine with delete vertex\n" );
      return false;
   }
}

void MU_DeleteFrameAnimFrame::undoRelease()
{
   FrameVertexList::iterator it;
   for ( it = m_list.begin(); it != m_list.end(); it++ )
   {
      for ( unsigned t = 0; t < (*it).list->size(); t++ )
      {
         (*(*it).list)[t]->release();
      }
      (*it).list->clear();
   }
}

unsigned MU_DeleteFrameAnimFrame::size()
{
   unsigned listSize = 0;
   FrameVertexList::iterator it;
   for ( it = m_list.begin(); it != m_list.end(); it++ )
   {
      listSize += sizeof( Model::FrameAnimVertexList ) + (*it).list->size() * sizeof( Model::FrameAnimVertex );
   }
   return sizeof(MU_DeleteFrameAnimFrame) + m_list.size() * sizeof(DeleteFrame) + listSize;
}

void MU_DeleteFrameAnimFrame::setAnimationData( const unsigned & anim )
{
   m_anim = anim;
}

void MU_DeleteFrameAnimFrame::deleteFrame( const unsigned & frame, Model::FrameAnimVertexList * list )
{
   DeleteFrame del;

   del.frame = frame;
   del.list  = list;

   m_list.push_back( del );
}

MU_SetVertexBoneJoint::MU_SetVertexBoneJoint()
{
}

MU_SetVertexBoneJoint::~MU_SetVertexBoneJoint()
{
}

void MU_SetVertexBoneJoint::undo( Model * model )
{
   log_debug( "undo set vertex bone joint\n" );
   SetJointList::iterator it;

   for ( it = m_list.begin(); it != m_list.end(); it++ )
   {
      model->setVertexBoneJoint( (*it).vertex, (*it).oldBone );
   }
}

void MU_SetVertexBoneJoint::redo( Model * model )
{
   SetJointList::iterator it;

   for ( it = m_list.begin(); it != m_list.end(); it++ )
   {
      model->setVertexBoneJoint( (*it).vertex, (*it).newBone );
   }
}

bool MU_SetVertexBoneJoint::combine( Undo * u )
{
   MU_SetVertexBoneJoint * undo = dynamic_cast< MU_SetVertexBoneJoint * >( u );

   if ( undo )
   {
      SetJointList::iterator it;

      for ( it = undo->m_list.begin(); it != undo->m_list.end(); it++ )
      {
         setVertexBoneJoint( (*it).vertex, (*it).newBone, (*it).oldBone );
      }

      return true;
   }
   else
   {
      log_debug( "couldn't combine with set vertex bone joint\n" );
      return false;
   }
}

unsigned MU_SetVertexBoneJoint::size()
{
   return sizeof(MU_SetVertexBoneJoint) + m_list.size() * sizeof(SetJoint);
}

void MU_SetVertexBoneJoint::setVertexBoneJoint( const unsigned & vertex, 
      const int & newBone, const int & oldBone )
{
   SetJointList::iterator it;

   for ( it = m_list.begin(); it != m_list.end(); it++ )
   {
      if ( (*it).vertex == vertex )
      {
         (*it).newBone = newBone;
         return;
      }
   }

   SetJoint sj;
   sj.vertex   = vertex;
   sj.newBone  = newBone;
   sj.oldBone  = oldBone;
   m_list.push_back( sj );
}

MU_AddAnimation::MU_AddAnimation()
   : m_skelAnim( NULL ),
     m_frameAnim( NULL )
{
}

MU_AddAnimation::~MU_AddAnimation()
{
}

void MU_AddAnimation::undo( Model * model )
{
   if ( m_skelAnim )
   {
      model->removeSkelAnim( m_anim );
      unsigned num = ( m_anim < model->getAnimCount( Model::ANIMMODE_SKELETAL ) ) ? m_anim : m_anim - 1;

      if ( model->getAnimationMode() != Model::ANIMMODE_NONE ) 
      {
         model->setCurrentAnimation( Model::ANIMMODE_SKELETAL, num );
      }
   }
   if ( m_frameAnim )
   {
      model->removeFrameAnim( m_anim );
      unsigned num = ( m_anim < model->getAnimCount( Model::ANIMMODE_FRAME ) ) ? m_anim : m_anim - 1;

      if ( model->getAnimationMode() != Model::ANIMMODE_NONE ) 
      {
         model->setCurrentAnimation( Model::ANIMMODE_FRAME, num );
      }
   }
}

void MU_AddAnimation::redo( Model * model )
{
   if ( m_skelAnim )
   {
      model->insertSkelAnim( m_anim, m_skelAnim );

      if ( model->getAnimationMode() != Model::ANIMMODE_NONE ) 
      {
         model->setCurrentAnimation( Model::ANIMMODE_SKELETAL, m_anim );
      }
   }
   if ( m_frameAnim )
   {
      model->insertFrameAnim( m_anim, m_frameAnim );

      if ( model->getAnimationMode() != Model::ANIMMODE_NONE ) 
      {
         model->setCurrentAnimation( Model::ANIMMODE_FRAME, m_anim );
      }
   }
}

bool MU_AddAnimation::combine( Undo * u )
{
   log_debug( "couldn't combine with add animation\n" );
   return false;
}

void MU_AddAnimation::redoRelease()
{
   log_debug("releasing animation in redo\n" );
   if ( m_skelAnim )
   {
      m_skelAnim->release();
   }
   if ( m_frameAnim )
   {
      m_frameAnim->release();
   }
}

unsigned MU_AddAnimation::size()
{
   unsigned frameAnimSize = (m_frameAnim) ? sizeof( m_frameAnim ) : 0;
   unsigned frame = 0;

   if ( m_frameAnim )
   {
      unsigned count = m_frameAnim->m_frameVertices.size();
      for ( frame = 0; frame < count; frame++ )
      {
         Model::FrameAnimVertexList * list = m_frameAnim->m_frameVertices[ frame ];
         frameAnimSize += sizeof( Model::FrameAnimVertexList ) + list->size() * sizeof( Model::FrameAnimVertex );
      }
   }

   unsigned skelAnimSize = (m_skelAnim) ? sizeof( m_skelAnim ) : 0;

   if ( m_skelAnim )
   {
      unsigned jointCount = m_skelAnim->m_jointKeyframes.size();
      for ( unsigned j = 0; j < jointCount; j++ )
      {
         Model::KeyframeList & list = m_skelAnim->m_jointKeyframes[j];
         skelAnimSize += sizeof( Model::KeyframeList ) + list.size() * sizeof( Model::Keyframe );
      }
      skelAnimSize += jointCount * sizeof( Model::JointKeyframeList );
   }

   return sizeof(MU_AddAnimation) + frameAnimSize + skelAnimSize;
}

void MU_AddAnimation::addAnimation( const unsigned & anim, Model::SkelAnim * skelanim )
{
   m_anim      = anim;
   m_skelAnim  = skelanim;
   m_frameAnim = NULL;
}

void MU_AddAnimation::addAnimation( const unsigned & anim, Model::FrameAnim * frameanim )
{
   m_anim      = anim;
   m_skelAnim  = NULL;
   m_frameAnim = frameanim;
}

MU_DeleteAnimation::MU_DeleteAnimation()
   : m_skelAnim( NULL ),
     m_frameAnim( NULL )
{
}

MU_DeleteAnimation::~MU_DeleteAnimation()
{
}

void MU_DeleteAnimation::undo( Model * model )
{
   if ( m_skelAnim )
   {
      model->insertSkelAnim( m_anim, m_skelAnim );
      if ( model->getAnimationMode() != Model::ANIMMODE_NONE ) 
      {
         model->setCurrentAnimation( Model::ANIMMODE_SKELETAL, m_anim );
      }
   }
   if ( m_frameAnim )
   {
      model->insertFrameAnim( m_anim, m_frameAnim );
      if ( model->getAnimationMode() != Model::ANIMMODE_NONE ) 
      {
         model->setCurrentAnimation( Model::ANIMMODE_FRAME, m_anim );
      }
   }
}

void MU_DeleteAnimation::redo( Model * model )
{
   if ( m_skelAnim )
   {
      model->removeSkelAnim( m_anim );
      unsigned num = ( m_anim < model->getAnimCount( Model::ANIMMODE_SKELETAL ) ) ? m_anim : m_anim - 1;
      if ( model->getAnimationMode() != Model::ANIMMODE_NONE ) 
      {
         model->setCurrentAnimation( Model::ANIMMODE_SKELETAL, num );
      }
   }
   if ( m_frameAnim )
   {
      model->removeFrameAnim( m_anim );
      unsigned num = ( m_anim < model->getAnimCount( Model::ANIMMODE_FRAME ) ) ? m_anim : m_anim - 1;
      if ( model->getAnimationMode() != Model::ANIMMODE_NONE ) 
      {
         model->setCurrentAnimation( Model::ANIMMODE_FRAME, num );
      }
   }
}

bool MU_DeleteAnimation::combine( Undo * u )
{
   log_debug( "couldn't combine with delete animation\n" );
   return false;
}

void MU_DeleteAnimation::undoRelease()
{
   log_debug("releasing animation in undo\n" );
   if ( m_skelAnim )
   {
      m_skelAnim->release();
   }
   if ( m_frameAnim )
   {
      m_frameAnim->release();
   }
}

unsigned MU_DeleteAnimation::size()
{
   unsigned frameAnimSize = (m_frameAnim) ? sizeof( m_frameAnim ) : 0;
   unsigned frame = 0;

   if ( m_frameAnim )
   {
      unsigned count = m_frameAnim->m_frameVertices.size();
      for ( frame = 0; frame < count; frame++ )
      {
         Model::FrameAnimVertexList * list = m_frameAnim->m_frameVertices[ frame ];
         frameAnimSize += sizeof( Model::FrameAnimVertexList ) + list->size() * sizeof( Model::FrameAnimVertex );
      }
   }

   unsigned skelAnimSize = (m_skelAnim) ? sizeof( m_skelAnim ) : 0;

   if ( m_skelAnim )
   {
      unsigned jointCount = m_skelAnim->m_jointKeyframes.size();
      for ( unsigned j = 0; j < jointCount; j++ )
      {
         Model::KeyframeList & list = m_skelAnim->m_jointKeyframes[j];
         skelAnimSize += sizeof( Model::KeyframeList ) + list.size() * sizeof( Model::Keyframe );
      }
      skelAnimSize += jointCount * sizeof( Model::JointKeyframeList );
   }

   return sizeof(MU_DeleteAnimation) + frameAnimSize + skelAnimSize;
}

void MU_DeleteAnimation::deleteAnimation( const unsigned & anim, Model::SkelAnim * skelanim )
{
   m_anim      = anim;
   m_skelAnim  = skelanim;
   m_frameAnim = NULL;
}

void MU_DeleteAnimation::deleteAnimation( const unsigned & anim, Model::FrameAnim * frameanim )
{
   m_anim      = anim;
   m_skelAnim  = NULL;
   m_frameAnim = frameanim;
}

MU_SetJointParent::MU_SetJointParent()
{
}

MU_SetJointParent::~MU_SetJointParent()
{
}

void MU_SetJointParent::undo( Model * model )
{
   model->setBoneJointParent( m_joint, m_oldParent );
}

void MU_SetJointParent::redo( Model * model )
{
   model->setBoneJointParent( m_joint, m_newParent );
}

void MU_SetJointParent::setJointParent( unsigned joint, int newParent, int oldParent )
{
   m_joint = joint;
   m_newParent = newParent;
   m_oldParent = oldParent;
}

MU_SetJointRotation::MU_SetJointRotation()
{
}

MU_SetJointRotation::~MU_SetJointRotation()
{
}

void MU_SetJointRotation::undo( Model * model )
{
   model->setBoneJointRotation( m_joint, m_oldRotation );
}

void MU_SetJointRotation::redo( Model * model )
{
   model->setBoneJointRotation( m_joint, m_newRotation );
}

void MU_SetJointRotation::setJointRotation( const unsigned & joint, 
      const double *  newRotation, const double * oldRotation )
{
   m_joint = joint;
   if ( newRotation )
   {
      for ( unsigned i = 0; i < 3; i++ )
      {
         m_newRotation[i] = newRotation[i];
      }
   }
   if ( oldRotation )
   {
      for ( unsigned i = 0; i < 3; i++ )
      {
         m_oldRotation[i] = oldRotation[i];
      }
   }
}

MU_SetJointTranslation::MU_SetJointTranslation()
{
}

MU_SetJointTranslation::~MU_SetJointTranslation()
{
}

void MU_SetJointTranslation::undo( Model * model )
{
   model->setBoneJointTranslation( m_joint, m_oldTranslation );
}

void MU_SetJointTranslation::redo( Model * model )
{
   model->setBoneJointTranslation( m_joint, m_newTranslation );
}

void MU_SetJointTranslation::setJointTranslation( const unsigned & joint, 
      const double *  newTranslation, const double * oldTranslation )
{
   m_joint = joint;
   if ( newTranslation )
   {
      for ( unsigned i = 0; i < 3; i++ )
      {
         m_newTranslation[i] = newTranslation[i];
      }
   }
   if ( oldTranslation )
   {
      for ( unsigned i = 0; i < 3; i++ )
      {
         m_oldTranslation[i] = oldTranslation[i];
      }
   }
}

MU_AddBoneJoint::MU_AddBoneJoint()
{
}

MU_AddBoneJoint::~MU_AddBoneJoint()
{
}

void MU_AddBoneJoint::undo( Model * model )
{
   model->removeBoneJoint( m_jointNum );
}

void MU_AddBoneJoint::redo( Model * model )
{
   model->insertBoneJoint( m_jointNum, m_joint );
}

bool MU_AddBoneJoint::combine( Undo * u )
{
   log_debug( "couldn't combine with add bone joint\n" );
   return false;
}

void MU_AddBoneJoint::redoRelease()
{
   m_joint->release();
}

unsigned MU_AddBoneJoint::size()
{
   return sizeof(MU_AddBoneJoint) + sizeof(Model::Joint);
}

void MU_AddBoneJoint::addBoneJoint( unsigned jointNum, Model::Joint * joint )
{
   m_jointNum = jointNum;
   m_joint    = joint;
}

MU_DeleteBoneJoint::MU_DeleteBoneJoint()
{
}

MU_DeleteBoneJoint::~MU_DeleteBoneJoint()
{
}

void MU_DeleteBoneJoint::undo( Model * model )
{
   log_debug( "undo delete joint\n" );
   model->insertBoneJoint( m_jointNum, m_joint );
}

void MU_DeleteBoneJoint::redo( Model * model )
{
   log_debug( "redo delete joint\n" );
   model->removeBoneJoint( m_jointNum );
}

bool MU_DeleteBoneJoint::combine( Undo * u )
{
   log_debug( "couldn't combine with delete bone joint\n" );
   return false;
}

void MU_DeleteBoneJoint::undoRelease()
{
   m_joint->release();
}

unsigned MU_DeleteBoneJoint::size()
{
   return sizeof(MU_DeleteBoneJoint) + sizeof(Model::Joint);
}

void MU_DeleteBoneJoint::deleteBoneJoint( unsigned jointNum, Model::Joint * joint )
{
   m_jointNum = jointNum;
   m_joint    = joint;
}

MU_SetGroupSmooth::MU_SetGroupSmooth()
{
}

MU_SetGroupSmooth::~MU_SetGroupSmooth()
{
}

void MU_SetGroupSmooth::undo( Model * model )
{
   SetSmoothList::iterator it;

   for ( it = m_list.begin(); it != m_list.end(); it++ )
   {
      model->setGroupSmooth( (*it).group, (*it).oldSmooth );
   }
}

void MU_SetGroupSmooth::redo( Model * model )
{
   SetSmoothList::iterator it;

   for ( it = m_list.begin(); it != m_list.end(); it++ )
   {
      model->setGroupSmooth( (*it).group, (*it).newSmooth );
   }
}

bool MU_SetGroupSmooth::combine( Undo * u )
{
   MU_SetGroupSmooth * undo = dynamic_cast< MU_SetGroupSmooth * >( u );

   if ( undo )
   {
      SetSmoothList::iterator it;

      for ( it = undo->m_list.begin(); it != undo->m_list.end(); it++ )
      {
         setGroupSmooth( (*it).group, (*it).newSmooth, (*it).oldSmooth );
      }

      return true;
   }
   else
   {
      log_debug( "couldn't combine with set smooth\n" );
      return false;
   }
}

unsigned MU_SetGroupSmooth::size()
{
   return sizeof(MU_SetGroupSmooth) + m_list.size() * sizeof(SetSmooth);
}

void MU_SetGroupSmooth::setGroupSmooth( const unsigned & group, 
      const uint8_t & newSmooth, const uint8_t & oldSmooth )
{
   SetSmoothList::iterator it;

   for ( it = m_list.begin(); it != m_list.end(); it++ )
   {
      if ( (*it).group == group )
      {
         (*it).newSmooth = newSmooth;
         return;
      }
   }

   SetSmooth ss;
   ss.group      = group;
   ss.newSmooth  = newSmooth;
   ss.oldSmooth  = oldSmooth;
   m_list.push_back( ss );
}

MU_SetGroupName::MU_SetGroupName()
{
}

MU_SetGroupName::~MU_SetGroupName()
{
}

void MU_SetGroupName::undo( Model * model )
{
   model->setGroupName( m_groupNum, m_oldName.c_str() );
}

void MU_SetGroupName::redo( Model * model )
{
   model->setGroupName( m_groupNum, m_newName.c_str() );
}

bool MU_SetGroupName::combine( Undo * u )
{
   MU_SetGroupName * undo = dynamic_cast< MU_SetGroupName * >( u );

   if ( undo && undo->m_groupNum == m_groupNum )
   {
      m_newName = undo->m_newName;
      return true;
   }
   else
   {
      return false;
   }
}

unsigned MU_SetGroupName::size()
{
   return sizeof(MU_SetGroupName);
}

void MU_SetGroupName::setGroupName( unsigned groupNum, const char * newName, const char * oldName )
{
   m_groupNum   = groupNum;
   m_newName    = newName;
   m_oldName    = oldName;
}

MU_MoveAnimation::MU_MoveAnimation()
{
}

MU_MoveAnimation::~MU_MoveAnimation()
{
}

void MU_MoveAnimation::undo( Model * model )
{
   model->moveAnimation( m_mode, m_newIndex, m_oldIndex );
}

void MU_MoveAnimation::redo( Model * model )
{
   model->moveAnimation( m_mode, m_oldIndex, m_newIndex );
}

bool MU_MoveAnimation::combine( Undo * u )
{
   return false;
}

unsigned MU_MoveAnimation::size()
{
   return sizeof(MU_MoveAnimation);
}

void MU_MoveAnimation::moveAnimation( const Model::AnimationMode & mode, 
      const unsigned & oldIndex, const unsigned & newIndex )
{
   m_mode     = mode;
   m_oldIndex = oldIndex;
   m_newIndex = newIndex;
}

MU_SetFrameAnimVertexCount::MU_SetFrameAnimVertexCount()
{
}

MU_SetFrameAnimVertexCount::~MU_SetFrameAnimVertexCount()
{
}

void MU_SetFrameAnimVertexCount::undo( Model * model )
{
   model->setFrameAnimVertexCount( m_oldCount );
}

void MU_SetFrameAnimVertexCount::redo( Model * model )
{
   model->setFrameAnimVertexCount( m_newCount );
}

unsigned MU_SetFrameAnimVertexCount::size()
{
   return sizeof(MU_SetFrameAnimVertexCount);
}

void MU_SetFrameAnimVertexCount::setCount( const unsigned & newCount, const unsigned & oldCount )
{
   m_newCount = newCount;
   m_oldCount = oldCount;
}

