/*  Misfit Model 3D
 * 
 *  Copyright (c) 2004 Kevin Worcester
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, 
 *  USA.
 *
 *  See the COPYING file for full license text.
 */


#include "pluginmgr.h"
#include "mm3dconfig.h"
#include "log.h"

#include <stdio.h>
#include <stdlib.h>
#include <dirent.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <sys/stat.h>

#ifdef HAVE_DLOPEN
#include <dlfcn.h>
#endif // HAVE_DLOPEN

#include <string>

using std::string;

PluginManager * PluginManager::s_instance = NULL;

PluginManager::PluginManager()
   : m_nextId( 0 ),
     m_initialize( true )
{
}

PluginManager::~PluginManager()
{
   unloadPlugins();
}

PluginManager * PluginManager::getInstance()
{
   if ( s_instance == NULL )
   {
      s_instance = new PluginManager();
   }

   return s_instance;
}

void PluginManager::release()
{
   if ( s_instance != NULL )
   {
      delete s_instance;
      s_instance = NULL;
   }
}

bool PluginManager::loadPlugin( const char * pluginFile )
{
   bool enabled = true;

#ifdef HAVE_DLOPEN
   void * handle;
   bool (*init_function)();
   bool (*uninit_function)();
   const char * (*version_function)();
   const char * (*desc_function)();
   PluginStatus status = PluginError;
   string name;

   log_debug( "loading plugin file: %s\n", pluginFile );
   name = fileToName( pluginFile );

   if( (handle = dlopen( pluginFile, RTLD_NOW | RTLD_GLOBAL )) == NULL )
   {
      log_warning( "dlopen: %s\n", dlerror() );
      return false;
   }

   if( (init_function = ( bool (*)()) dlsym(handle, "plugin_init")) == NULL)
   {
      log_warning( "%s: no plugin_init symbol\n", pluginFile );
      enabled = false;
      status = PluginNotPlugin;
   }

   if( (uninit_function = ( bool (*)()) dlsym(handle, "plugin_uninit")) == NULL)
   {
      log_warning( "%s: no plugin_uninit symbol\n", pluginFile );
      status = PluginNotPlugin;
      enabled = false;
   }

   if( (version_function = ( const char * (*)()) dlsym(handle, "plugin_version")) == NULL)
   {
      // This is just a warning, don't fail over it
      log_warning( "%s: no plugin_version symbol\n", pluginFile );
   }

   if( (desc_function = ( const char * (*)()) dlsym(handle, "plugin_desc")) == NULL)
   {
      // This is just a warning, don't fail over it
      log_warning( "%s: no plugin_desc symbol\n", pluginFile );
   }

   // Try to initialize
   if ( m_initialize && !isDisabled( name.c_str() ) )
   { 
      if ( init_function && init_function() )
      {
         status = PluginActive;
         enabled = true;
      }
      else
      {
         status = PluginError;
         enabled = false;
         log_error( "initialization failed for %s\n", pluginFile );
         dlclose( handle );
      }
   }
   else
   {
      log_debug( "%s disabled\n", pluginFile );
      status = PluginDisabled;
      enabled = false;
   }

   PluginData * data = new PluginData();

   data->m_id              = getNextId();
   data->m_status          = status;
   data->m_name            = name;
   data->m_enabled         = enabled;
   data->m_initFunction    = init_function;
   data->m_uninitFunction  = uninit_function;
   data->m_versionFunction = version_function;
   data->m_descFunction    = desc_function;
   data->m_fileHandle      = handle;

   m_plugins.push_back( data );
#else
   log_warning( "plugins disabled at compile time\n" );
#endif // HAVE_DLOPEN

   return enabled;
}

bool PluginManager::loadPluginDir( const char * pluginDir )
{
   if ( pluginDir && pluginDir[0] )
   {
      DIR * dp = opendir( pluginDir );
      if ( dp )
      {
         log_debug( "loading plugin directory: %s\n", pluginDir );

         struct dirent * d;
         while ( (d = readdir( dp ) ) != NULL )
         {
            if ( strcmp( d->d_name, "." ) == 0 || strcmp( d->d_name, "..") == 0 )
            {
               continue;
            }
            string file = string(pluginDir) + DIR_SLASH;
            file += d->d_name;

            struct stat statbuf;
            if ( lstat( file.c_str(), &statbuf ) == 0 )
            {
               if ( S_ISREG( statbuf.st_mode ) )
               {
                  loadPlugin( file.c_str() );
               }
               else if ( S_ISLNK( statbuf.st_mode ) )
               {
                  log_debug( "%s is a symlink\n", file.c_str() );
                  // Try directory, then regular file
                  if ( ! loadPluginDir( file.c_str() ) )
                  {
                     // Not a directory, let's try file
                     loadPlugin( file.c_str() );
                  }
               }
               else if ( S_ISDIR( statbuf.st_mode ) )
               {
                  loadPluginDir( file.c_str() );
               }
               else
               {
                  log_warning( "Don't know what to do with st_mode %08X\n", 
                        (unsigned) statbuf.st_mode );
               }
            }
            else
            {
               log_error( "stat: %s: %s\n", file.c_str(), strerror(errno) );
            }
         }

         closedir( dp );
         return true;
      }
      else
      {
         // This is called any time the file is not a regular file
         // If it turns out to not be a directory, that isn't really
         // an error.
         if ( errno != ENOTDIR )
         {
            log_warning( "%s: %s\n", pluginDir, strerror(errno) );
         }
      }
   }

   return false;
}

bool PluginManager::unloadPlugins()
{
   log_debug( "PluginManager unloading %d plugins\n", m_plugins.size() );
#ifdef HAVE_DLOPEN
   PluginDataList::iterator it = m_plugins.begin();

   while ( it != m_plugins.end() )
   {
      if ( (*it)->m_enabled )
      {
         (*it)->m_uninitFunction();
         dlclose( (*it)->m_fileHandle );
      }
      delete *it;
      it++;
   }

   m_plugins.clear();
#else
   log_warning( "plugins disabled at compile time\n" );
#endif // HAVE_DLOPEN

   return true;
}

void PluginManager::loadPlugins()
{
   string plugin_dir = getenv("HOME");
   plugin_dir += HOME_PLUGINS;

   // Check to see if we have a plugin directory in HOME
   if ( ! loadPluginDir( plugin_dir.c_str() ) )
   {
      // Not in home, check shared directory
      plugin_dir  = SHARED_PLUGINS;
      loadPluginDir( plugin_dir.c_str() );
   }
}

int init_plugins()
{
   PluginManager * mgr = PluginManager::getInstance();
   mgr->loadPlugins();
   return 0;
}

int uninit_plugins()
{
   PluginManager * mgr = PluginManager::getInstance();
   mgr->unloadPlugins();
   return 0;
}

list<int> PluginManager::getPluginIds()
{
   list<int> pluginList;

   PluginDataList::iterator it;
   for ( it = m_plugins.begin(); it != m_plugins.end(); it++ )
   {
      pluginList.push_back( (*it)->m_id );
   }

   return pluginList;
}

const char * PluginManager::getPluginName( int id )
{
   PluginDataList::iterator it;
   for ( it = m_plugins.begin(); it != m_plugins.end(); it++ )
   {
      if ( (*it)->m_id == id )
      {
         return (*it)->m_name.c_str();
      }
   }

   return "Not found";
}

PluginManager::PluginStatus PluginManager::getPluginStatus( int id )
{
   PluginDataList::iterator it;
   for ( it = m_plugins.begin(); it != m_plugins.end(); it++ )
   {
      if ( (*it)->m_id == id )
      {
         return (*it)->m_status;
      }
   }

   return PluginError;
}

const char * PluginManager::getPluginVersion( int id )
{
   PluginDataList::iterator it;
   for ( it = m_plugins.begin(); it != m_plugins.end(); it++ )
   {
      if ( (*it)->m_id == id )
      {
         if ( (*it)->m_versionFunction )
         {
            const char * str = (*it)->m_versionFunction();

            return str ? str : "Unavailable";
         }
         else
         {
            return "Unavailable";
         }
      }
   }

   return "Not found";
}

const char * PluginManager::getPluginDescription( int id )
{
   PluginDataList::iterator it;
   for ( it = m_plugins.begin(); it != m_plugins.end(); it++ )
   {
      if ( (*it)->m_id == id )
      {
         if ( (*it)->m_descFunction )
         {
            const char * str = (*it)->m_descFunction();

            return str ? str : "Unavailable";
         }
         else
         {
            return "Unavailable";
         }
      }
   }

   return "Not found";
}

string PluginManager::fileToName( const char * file )
{
   char * temp = strdup( file );
   char * s = strrchr( temp, DIR_SLASH );
   if ( s )
   {
      s += 1; // skip last slash
   }
   else
   {
      s = temp; // no slash, just start with filename
   }

   int len = (int) strlen( s );
   for ( int t = len - 1; t >= 0; t-- )
   {
      if ( s[t] == '.' )
      {
         s[t] = '\0';
         break;
      }
   }

   string str = s;
   free( temp );

   return str;
}

void PluginManager::disable( const char * pluginName )
{
   list<string>::iterator it;

   for ( it = m_disabled.begin(); it != m_disabled.end(); it++ )
   {
      if ( strcasecmp ( (*it).c_str(), pluginName  ) == 0 )
      {
         return;
      }
   }

   m_disabled.push_back( pluginName );
}

bool PluginManager::isDisabled( const char * pluginName )
{
   list<string>::iterator it;

   for ( it = m_disabled.begin(); it != m_disabled.end(); it++ )
   {
      if ( strcasecmp ( (*it).c_str(), pluginName  ) == 0 )
      {
         return true;
      }
   }

   return false;
}
