/*  Misfit Model 3D
 * 
 *  Copyright (c) 2004 Kevin Worcester
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, 
 *  USA.
 *
 *  See the COPYING file for full license text.
 */


#include "qttex.h"

#include "mm3dconfig.h"
#include "log.h"

#include <qimage.h>
#include <qstring.h>

using std::list;
using std::string;

QtTextureFilter::QtTextureFilter()
{
   QStrList list;
   char * str;

   list = QImage::inputFormats();
   for ( str = list.first(); str; str = list.next() )
   {
      m_read.push_back( str );
      if ( strcasecmp( str, "JPEG" ) == 0 )
      {
         m_read.push_back( "JPG" );
      }
      if ( strcasecmp( str, "TIFF" ) == 0 )
      {
         m_read.push_back( "TIF" );
      }
   }

   list = QImage::outputFormats();
   for ( str = list.first(); str; str = list.next() )
   {
      m_write.push_back( str );
      if ( strcasecmp( str, "JPEG" ) == 0 )
      {
         m_write.push_back( "JPG" );
      }
      if ( strcasecmp( str, "TIFF" ) == 0 )
      {
         m_write.push_back( "TIF" );
      }
   }
}

QtTextureFilter::~QtTextureFilter()
{
}

bool QtTextureFilter::canRead( const char * filename )
{
   if ( filename == NULL )
   {
      return false;
   }

   QString cmpstr;
   unsigned len = strlen( filename );

   QStringList::Iterator it;
   
   for ( it = m_read.begin(); it != m_read.end(); it++ )
   {
      cmpstr = QString(".") + *it;

      if ( len >= cmpstr.length() )
      {
         if ( strcasecmp( &filename[len-cmpstr.length()], cmpstr.latin1()) == 0 )
         {
            return true;
         }
      }
   }

   return false;
}

Texture::Error QtTextureFilter::readFile( Texture * texture, const char * filename )
{
   QImage image;

   if ( filename == NULL )
   {
      log_error( "filename NULL\n" );
      return Texture::ERROR_NO_FILE;
   }

   if ( texture == NULL )
   {
      log_error( "texture NULL\n" );
      return Texture::ERROR_UNKNOWN;
   }

   if ( image.load( filename ) )
   {
      texture->m_width  = image.width();
      texture->m_height = image.height();

      bool hasAlpha = image.hasAlphaBuffer();
      log_debug( "Alpha channel: %s\n", hasAlpha ? "present" : "not present" );

      unsigned pixelBytes = hasAlpha ? 4 : 3;
      unsigned pixelCount = texture->m_width * texture->m_height;
      unsigned imageSize = pixelCount * (pixelBytes * sizeof(uint8_t));
      texture->m_data = new uint8_t[ imageSize ];

      // Make bottom row the first row, as required by OpenGL
      if ( hasAlpha )
      {
         texture->m_format = Texture::FORMAT_RGBA;
         for ( int y = 0; y < texture->m_height; y ++ )
         {
            for ( int x = 0; x < texture->m_width; x++ )
            {
               QRgb p = image.pixel( x, texture->m_height - y - 1 );
               texture->m_data[ ((y * texture->m_width + x)*4) + 0 ] = qRed( p );
               texture->m_data[ ((y * texture->m_width + x)*4) + 1 ] = qGreen( p );
               texture->m_data[ ((y * texture->m_width + x)*4) + 2 ] = qBlue( p );
               texture->m_data[ ((y * texture->m_width + x)*4) + 3 ] = qAlpha( p );
            }
         }
      }
      else
      {
         int tmax = 0;
         texture->m_format = Texture::FORMAT_RGB;
         for ( int y = 0; y < texture->m_height; y ++ )
         {
            for ( int x = 0; x < texture->m_width; x++ )
            {
               QRgb p = image.pixel( x, texture->m_height - y - 1 );
               texture->m_data[ ((y * texture->m_width + x)*3) + 0 ] = qRed( p );
               texture->m_data[ ((y * texture->m_width + x)*3) + 1 ] = qGreen( p );
               texture->m_data[ ((y * texture->m_width + x)*3) + 2 ] = qBlue( p );

               tmax = ((y * texture->m_width + x)*3) + 2;
            }
         }
         log_debug( "last read = %d\n", tmax );
      }

      texture->m_filename = strdup( filename );

      const char * name = strrchr( filename, DIR_SLASH );
      if ( name )
      {
         texture->m_name = strdup( &name[1] );
      }
      else
      {
         texture->m_name = strdup( filename );
      }
      char * ext = strrchr( texture->m_name, '.' );
      if ( ext )
      {
         ext[0] = '\0';
      }

      return Texture::ERROR_NONE;
   }
   else
   {
      return Texture::ERROR_FILE_OPEN;
   }

   return Texture::ERROR_UNKNOWN;
}

list<string> QtTextureFilter::getReadTypes()
{
   list<string> rval;

   QStringList::Iterator it;
   for ( it = m_read.begin(); it != m_read.end(); it++ )
   {
      rval.push_back( (QString("*.") + *it).latin1() );
   }

   return rval;
}

list<string> QtTextureFilter::getWriteTypes()
{
   list<string> rval;

   QStringList::Iterator it;
   for ( it = m_write.begin(); it != m_write.end(); it++ )
   {
      rval.push_back( (QString("*.") + *it).latin1() );
   }

   return rval;
}


