/*  Misfit Model 3D
 * 
 *  Copyright (c) 2004 Kevin Worcester
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, 
 *  USA.
 *
 *  See the COPYING file for full license text.
 */


#include "texwidget.h"

#include "texture.h"
#include "log.h"
#include <math.h>

using std::vector;
using std::list;

TextureWidget::TextureWidget( QWidget * parent, const char * name )
   : QGLWidget( parent, name ),
     m_texture( NULL ),
     m_glTexture( 0 ),
     m_selecting( false ),
     m_drawBounding( false )
{
}

TextureWidget::~TextureWidget()
{
   makeCurrent();
   glDeleteTextures( 1, (GLuint *) &m_glTexture );
   // Do NOT free m_texture.  TextureManager does that.
}

void TextureWidget::initializeGL()
{
   glEnable( GL_TEXTURE_2D );

   glShadeModel( GL_SMOOTH );
   glClearColor( 0.80, 0.80, 0.80, 1.0 );
   glClearDepth( 1.0f );

   glGenTextures( 1, &m_glTexture );
}

void TextureWidget::resizeGL( int w, int h )
{
   if ( h == 0 )
      h = 1;

   m_viewportWidth  = w;
   m_viewportHeight = h;

   glViewport( 0, 0, ( GLint ) m_viewportWidth, ( GLint ) m_viewportHeight );

   glMatrixMode( GL_PROJECTION );
   glLoadIdentity( );

   GLfloat ratio = ( GLfloat ) m_viewportWidth / ( GLfloat ) m_viewportHeight;

   glPolygonMode( GL_FRONT_AND_BACK, GL_FILL );

   float x = 1.0;
   float y = x / ratio;
   glOrtho( 0.0, 1.0, 0.0, 1.0, -1.0, 1.0 );

   m_xOffset = x;
   m_yOffset = y;

   glMatrixMode( GL_MODELVIEW );
   glLoadIdentity( );

   updateGL();
}

void TextureWidget::paintGL()
{
   glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
   glLoadIdentity( );

   if ( m_texture )
   {
      glEnable( GL_TEXTURE_2D );

      glPolygonMode( GL_FRONT_AND_BACK, GL_FILL );

      glBegin( GL_QUADS );

      glColor3f( 1.0, 1.0, 1.0 );

      glBindTexture( GL_TEXTURE_2D, m_glTexture );

      glTexCoord2f( 0.0, 0.0 );
      glNormal3f( 0.0, 0.0, -1.0 );
      glVertex3f( 0.0, 0.0, 0.0 );

      glTexCoord2f( 1.0, 0.0 );
      glNormal3f( 0.0, 0.0, -1.0 );
      glVertex3f(  1.0, 0.0, 0.0 );

      glTexCoord2f( 1.0, 1.0 );
      glNormal3f( 0.0, 0.0, -1.0 );
      glVertex3f(  1.0,  1.0, 0.0 );

      glTexCoord2f( 0.0, 1.0 );
      glNormal3f( 0.0, 0.0, -1.0 );
      glVertex3f( 0.0,  1.0, 0.0 );

      glEnd();

      glDisable( GL_TEXTURE_2D );

      glColor3f( 1.0, 1.0, 1.0 );

      glPolygonMode( GL_FRONT_AND_BACK, GL_LINE );

      glBegin( GL_TRIANGLES );

      for ( unsigned t = 0; t < m_triangles.size(); t++ )
      {
         TextureTriangle * triangle = m_triangles[t];

         for ( unsigned v = 0; v < 3; v++ )
         {
            glVertex3f( m_vertices[ triangle->vertex[v] ]->s, 
                  m_vertices[ triangle->vertex[v] ]->t, 
                  -0.5 );
         }
      }

      glEnd();

      glPointSize( 3.0 );

      glBegin( GL_POINTS );

      for ( unsigned t = 0; t < m_triangles.size(); t++ )
      {
         for ( unsigned v = 0; v < 3; v++ )
         {
            TextureVertex * vert = m_vertices[ m_triangles[t]->vertex[v] ];
            if ( vert->selected )
            {
               glColor3f( 1.0, 0.0, 0.0 );
            }
            else
            {
               glColor3f( 1.0, 1.0, 1.0 );
            }

            glVertex3f( vert->s, vert->t, -0.5 );
         }
      }

      glEnd();

      if ( m_drawBounding )
      {
         drawSelectBox();
      }
   }
   else
   {
      glBegin( GL_QUADS );

      glColor3f( 0.0, 0.0, 0.0 );

      glVertex3f( -1.0, -1.0, 0.0 );
      glVertex3f(  1.0, -1.0, 0.0 );
      glVertex3f(  1.0,  1.0, 0.0 );
      glVertex3f( -1.0,  1.0, 0.0 );
      
      glEnd();
   }


   swapBuffers();
}

void TextureWidget::setTexture( Texture * texture )
{
   m_texture = texture;

   if ( m_texture )
   {
      makeCurrent();

      glBindTexture( GL_TEXTURE_2D, m_glTexture );

      glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER,
            GL_NEAREST );
      glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER,
            GL_NEAREST );

      GLuint format = m_texture->m_format == Texture::FORMAT_RGBA ? GL_RGBA : GL_RGB;
      gluBuild2DMipmaps( GL_TEXTURE_2D, format,
            m_texture->m_width, m_texture->m_height,
            format, GL_UNSIGNED_BYTE,
            m_texture->m_data );
   }

   updateGL();
}

int TextureWidget::addVertex( double s, double t )
{
   int index = m_vertices.size();

   TextureVertex * vert = new TextureVertex;
   vert->s = s;
   vert->t = t;
   vert->selected = true;
   m_vertices.push_back( vert );

   return index;
}

int TextureWidget::addTriangle( int v1, int v2, int v3 )
{
   int vCount = m_vertices.size();

   if (     v1 >= 0 && v1 < vCount
         && v2 >= 0 && v2 < vCount
         && v3 >= 0 && v3 < vCount )
   {
      int index = m_triangles.size();

      TextureTriangle * triangle = new TextureTriangle;
      triangle->vertex[0] = v1;
      triangle->vertex[1] = v2;
      triangle->vertex[2] = v3;
      m_triangles.push_back( triangle );

      return index;
   }

   return -1;
}

void TextureWidget::mousePressEvent( QMouseEvent * e )
{
   m_lastXPos = e->pos().x();
   m_lastYPos = e->pos().y();

   switch ( m_operation )
   {
      case MouseSelect:
         if ( !( (e->state() & ShiftButton) || (e->button() &  RightButton) ) )
         {
            clearSelected();
         }
         m_xSel1 =  m_lastXPos / (double) this->width();
         m_ySel1 =  1.0 - (m_lastYPos / (double) this->height());
         m_selecting = ( e->button() & RightButton ) ? false : true;
         m_drawBounding = true;
         break;
      case MouseMove:
         break;
      case MouseScale:
         startScale(
               m_lastXPos / (double) this->width(), 
               1.0 - (m_lastYPos / (double) this->height()) );
         break;
      default:
         log_error( "Unknown mouse operation: %d\n", m_operation );
         break;
   }
}

void TextureWidget::mouseReleaseEvent( QMouseEvent * e )
{
   int x = e->pos().x();
   int y = e->pos().y();

   switch ( m_operation )
   {
      case MouseSelect:
         m_drawBounding = false;
         updateSelectRegion( 
               x / (double) this->width(), 
               1.0 - (y / (double) this->height()) );
         selectDone();
         break;
      case MouseMove:
         moveSelectedVertices( 
               (x - m_lastXPos) / (double) this->width(), 
               -(y - m_lastYPos) / (double) this->height() );
         emit updateCoordinatesSignal();
         break;
      case MouseScale:
         break;
      default:
         log_error( "Unknown mouse operation: %d\n", m_operation );
         break;
   }
}

void TextureWidget::mouseMoveEvent( QMouseEvent * e )
{
   int x = e->pos().x();
   int y = e->pos().y();

   switch ( m_operation )
   {
      case MouseSelect:
         updateSelectRegion( 
               x / (double) this->width(), 
               1.0 - (y / (double) this->height()) );
         break;
      case MouseMove:
         moveSelectedVertices( 
               (x - m_lastXPos) / (double) this->width(), 
               -(y - m_lastYPos) / (double) this->height() );
         emit updateCoordinatesSignal();
         break;
      case MouseScale:
         scaleSelectedVertices(
               x / (double) this->width(), 
               1.0 - (y / (double) this->height()) );
         emit updateCoordinatesSignal();
         break;
      default:
         log_error( "Unknown mouse operation: %d\n", m_operation );
         break;
   }

   m_lastXPos = x;
   m_lastYPos = y;
}

void TextureWidget::moveSelectedVertices( double x, double y )
{
   for ( unsigned t = 0; t < m_vertices.size(); t++ )
   {
      if ( m_vertices[t]->selected )
      {
         m_vertices[t]->s += x;
         m_vertices[t]->t += y;
      }
   }
   updateGL();
}

void TextureWidget::updateSelectRegion( double x, double y )
{
   m_xSel2 = x;
   m_ySel2 = y;
   updateGL();
}

void TextureWidget::selectDone()
{
   if ( m_xSel1 > m_xSel2 )
   {
      double temp = m_xSel2;
      m_xSel2 = m_xSel1;
      m_xSel1 = temp;
   }
   if ( m_ySel1 > m_ySel2 )
   {
      double temp = m_ySel2;
      m_ySel2 = m_ySel1;
      m_ySel1 = temp;
   }
   for ( unsigned v = 0; v < m_vertices.size(); v++ )
   {
      if (     m_vertices[v]->s >= m_xSel1 && m_vertices[v]->s <= m_xSel2 
            && m_vertices[v]->t >= m_ySel1 && m_vertices[v]->t <= m_ySel2 )
      {
         m_vertices[v]->selected = m_selecting;
      }
   }
   updateGL();
}

void TextureWidget::setMouseOperation( MouseOperation op )
{
   log_debug( "texture widget operation = %d\n", (int) op);
   m_operation = op;
}

void TextureWidget::drawSelectBox()
{
   glEnable( GL_COLOR_LOGIC_OP );
   glColor3f( 1.0, 1.0, 1.0 );
   glLogicOp( GL_XOR );
   glBegin( GL_LINES );

   glVertex3f( m_xSel1, m_ySel1, -0.75 );
   glVertex3f( m_xSel1, m_ySel2, -0.75 );

   glVertex3f( m_xSel1, m_ySel2, -0.75 );
   glVertex3f( m_xSel2, m_ySel2, -0.75 );

   glVertex3f( m_xSel2, m_ySel2, -0.75 );
   glVertex3f( m_xSel2, m_ySel1, -0.75 );

   glVertex3f( m_xSel2, m_ySel1, -0.75 );
   glVertex3f( m_xSel1, m_ySel1, -0.75 );

   glEnd();
   glLogicOp( GL_COPY );
   glDisable( GL_LOGIC_OP );
}

void TextureWidget::clearSelected()
{
   for ( unsigned v = 0; v < m_vertices.size(); v++ )
   {
      m_vertices[v]->selected = false;
   }
}

void TextureWidget::clearCoordinates()
{
   while ( m_triangles.size() )
   {
      delete m_triangles.back();
      m_triangles.pop_back();
   }

   while ( m_vertices.size() )
   {
      delete m_vertices.back();
      m_vertices.pop_back();
   }
}

void TextureWidget::getCoordinates( int tri, float * s, float * t )
{
   if ( !t || !s || tri >= (signed) m_triangles.size() )
   {
      return;
   }

   for ( int v = 0; v < 3; v++ )
   {
      s[v] = m_vertices[ m_triangles[ tri ]->vertex[v] ]->s;
      t[v] = m_vertices[ m_triangles[ tri ]->vertex[v] ]->t;
   }
}

void TextureWidget::startScale( double x, double y )
{
   log_debug( "starting scale at %f,%f\n", x, y );

   m_scaleList.clear();
   bool first = true;

   double minX = 0;
   double minY = 0;
   double maxX = 0;
   double maxY = 0;

   for ( unsigned t = 0; t < m_vertices.size(); t++ )
   {
      if ( m_vertices[t]->selected )
      {
         // update range
         if ( first )
         {
            minX = m_vertices[t]->s;
            minY = m_vertices[t]->t;
            maxX = m_vertices[t]->s;
            maxY = m_vertices[t]->t;

            first = false;
         }
         else
         {
            if ( m_vertices[t]->s < minX ) { minX = m_vertices[t]->s; };
            if ( m_vertices[t]->t < minY ) { minY = m_vertices[t]->t; };
            if ( m_vertices[t]->s > maxX ) { maxX = m_vertices[t]->s; };
            if ( m_vertices[t]->t > maxY ) { maxY = m_vertices[t]->t; };
         }

         ScaleVertices sv;
         sv.index = t;
         sv.x = m_vertices[t]->s;
         sv.y = m_vertices[t]->t;

         m_scaleList.push_back( sv );
      }
   }

   double minmin = distance( x, y, minX, minY );
   double minmax = distance( x, y, minX, maxY );
   double maxmin = distance( x, y, maxX, minY );
   double maxmax = distance( x, y, maxX, maxY );

   if ( minmin > minmax )
   {
      if ( minmin > maxmin )
      {
         if ( minmin > maxmax )
         {
            m_farX = minX;
            m_farY = minY;
         }
         else
         {
            m_farX = maxX;
            m_farY = maxY;
         }
      }
      else // maxmin > minmin
      {
         if ( maxmin > maxmax )
         {
            m_farX = maxX;
            m_farY = minY;
         }
         else
         {
            m_farX = maxX;
            m_farY = maxY;
         }
      }
   }
   else // minmax > minmin
   {
      if ( minmax > maxmin )
      {
         if ( minmax > maxmax )
         {
            m_farX = minX;
            m_farY = maxY;
         }
         else
         {
            m_farX = maxX;
            m_farY = maxY;
         }
      }
      else // maxmin > minmax
      {
         if ( maxmin > maxmax )
         {
            m_farX = maxX;
            m_farY = minY;
         }
         else
         {
            m_farX = maxX;
            m_farY = maxY;
         }
      }
   }

   m_startLengthX = fabs( x - m_farX );
   m_startLengthY = fabs( y - m_farY );
   log_debug( "far distance is %f,%f\n", m_startLengthX, m_startLengthY );
}

void TextureWidget::scaleSelectedVertices( double x, double y )
{
   log_debug( "scaling at %f,%f\n", x, y );

   double lengthX = fabs( x - m_farX );
   double lengthY = fabs( y - m_farY );

   log_debug( "far distance is %f,%f\n", lengthX, lengthY );

   ScaleVerticesList::iterator it;
   for( it = m_scaleList.begin(); it != m_scaleList.end(); it++ )
   {
      double x = (*it).x;
      double y = (*it).y;

      x -= m_farX;
      y -= m_farY;

      x *= (lengthX / m_startLengthX);
      y *= (lengthY / m_startLengthY);

      x += m_farX;
      y += m_farY;

      m_vertices[(*it).index]->s = x;
      m_vertices[(*it).index]->t = y;
   }

   updateGL();
}

double TextureWidget::distance( const double & x1, const double & y1, const double & x2, const double & y2 )
{
   double xDiff = x2 - x1;
   double yDiff = y2 - y1;
   return sqrt( xDiff*xDiff + yDiff*yDiff );
}

double TextureWidget::max( const double & a, const double & b )
{
   return ( a > b ) ? a : b;
}

